import argparse
import os
import pathlib
import re
import shutil
import subprocess
import sys
import textwrap

import click
import numpy as np
import prettytable

from . import dataset
from . import local
from . import output
from . import rsync
from . import scp
from . import search
from . import ssh
from . import yaml
from ._version import version
from .external import exec_cmd

# set filename defaults
f_hostinfo = "shelephant_hostinfo.yaml"
f_dump = "shelephant_dump.yaml"


def _shelephant_parse_parser():
    """
    Return parser for :py:func:`shelephant_parse`.
    """

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    desc = "Parse a YAML-file, and print to screen."
    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("file", type=pathlib.Path, help="File path.")
    return parser


def shelephant_parse(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _shelephant_parse_parser()
    args = parser.parse_args(args)
    data = yaml.read(args.file)
    yaml.preview(data)


def _shelephant_dump_parser():
    """
    Return parser for :py:func:`shelephant_dump`.
    """

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    desc = textwrap.dedent(
        """\
    Dump filenames to a YAML-file.

    .. note::

        If you have too many arguments you can hit the pipe-limit. In that case, use ``xargs``:

        .. code-block:: bash

            find . -name "*.py" | xargs shelephant_dump -o dump.yaml

        or you can use ``--command`` such that *shelephant* executes the command for you:

        .. code-block:: bash

            shelephant_dump -o dump.yaml --command find . -name '*.py'
    """
    )

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument(
        "-o", "--output", type=pathlib.Path, default=f_dump, help="Output YAML-file."
    )
    parser.add_argument("--search", type=pathlib.Path, help='Run "search" in "root".')
    parser.add_argument("-a", "--append", action="store_true", help="Append existing file.")
    parser.add_argument("-i", "--info", action="store_true", help="Add information (sha256, size).")
    parser.add_argument(
        "-e", "--exclude", type=str, action="append", help="Exclude input matching this pattern."
    )
    parser.add_argument(
        "-E",
        "--exclude-extension",
        type=str,
        action="append",
        default=[],
        help='Exclude input with this extension (e.g. ".bak").',
    )
    parser.add_argument(
        "-k", "--keep", type=str, action="append", help="Keep only input matching this regex."
    )
    parser.add_argument("--fmt", type=str, help='Formatter of each line, e.g. ``"mycmd {}"``.')
    parser.add_argument(
        "-c",
        "--command",
        action="store_true",
        help="Interpret arguments as a command (instead of as filenames) an run it.",
    )
    parser.add_argument("--cwd", type=str, help="Directory to run ``--command``.")
    parser.add_argument(
        "--root", type=str, help="Root for relative paths (default: directory of output file)."
    )
    parser.add_argument("--abspath", action="store_true", help="Store as absolute paths.")
    parser.add_argument("-s", "--sort", action="store_true", help="Sort filenames.")
    parser.add_argument(
        "-f", "--force", action="store_true", help="Overwrite output file without prompt."
    )
    parser.add_argument("-v", "--version", action="version", version=version, help="")
    parser.add_argument("files", type=str, nargs="*", help="Filenames.")

    return parser


def shelephant_dump(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _shelephant_dump_parser()
    args = parser.parse_args(args)
    files = args.files
    assert not (args.command and args.cwd is not None), "Cannot use --cwd without --command."

    if args.search:
        assert len(files) == 0, "Cannot use both --search and filenames."
        assert args.root is None, "Root inferred from --search."
        assert not args.command, "Cannot use both --search and --command."
        loc = dataset.Location.from_yaml(args.search)
        loc.read()
        if args.info:
            loc.getinfo()
        root = loc.root
        files = loc.files(info=args.info)
    else:
        assert len(files) > 0, "Nothing to dump."
        if args.root:
            root = args.root
        else:
            root = args.output.parent

        if args.command:
            cmd = " ".join(files)
            files = subprocess.check_output(cmd, shell=True, cwd=args.cwd).decode("utf-8")
            files = list(filter(None, files.splitlines()))
            if args.cwd is not None:
                files = [os.path.join(args.cwd, file) for file in files]

    if args.abspath:
        files = [os.path.abspath(file) for file in files]
    elif args.search is None:
        files = [os.path.relpath(file, root) for file in files]

    if args.keep:
        ret = []
        for pattern in args.keep:
            ret += [file for file in files if re.match(pattern, file)]
        files = ret

    if args.exclude:
        excl = np.zeros(len(files), dtype=bool)
        for pattern in args.exclude:
            excl = np.logical_or(excl, np.array([re.match(pattern, file) for file in files]))
        files = [file for file, ex in zip(files, excl) if not ex]

    if args.exclude_extension:
        files = [file for file in files if pathlib.Path(file).suffix not in args.exclude_extension]

    if args.sort:
        files = sorted(files)

    if args.fmt:
        files = [args.fmt.format(file) for file in files]

    if args.info and not args.search:
        files = dataset.Location(root=root, files=files).getinfo().files(info=True)

    if args.append:
        main = yaml.read(args.output)
        assert type(main) == list, 'Can only append a "flat" file'
        files = main + files
        args.force = True

    yaml.dump(args.output, files, args.force)


def _shelephant_cp_parser():
    """
    Return parser for :py:func:`shelephant_cp`.
    """

    desc = textwrap.dedent(
        """
        Copy files listed in a (field of a) YAML-file.
        These filenames are assumed either relative to the YAML-file or absolute.

        Usage::

            shelephant_cp <sourceinfo.yaml> <dest_dirname>
            shelephant_cp <sourceinfo.yaml> <dest_dirname_on_host> --ssh <user@host>
            shelephant_cp <sourceinfo.yaml> <destinfo.yaml>

        .. note::

            Files that are marked as equal are not touched.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("--ssh", type=str, help="SSH destination (e.g. user@host).")
    parser.add_argument("--colors", type=str, default="dark", help="Color scheme [none, dark].")
    parser.add_argument(
        "--mode",
        type=str,
        help="Use 'sha256', 'rsync', and/or 'basic'.",
        default="sha256,rsync" if shutil.which("rsync") is not None else "sha256,basic",
    )
    parser.add_argument("-f", "--force", action="store_true", help="Overwrite without prompt.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("-n", "--dry-run", action="store_true", help="Print copy-plan and exit.")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("source", type=pathlib.Path, help="Source information.")
    parser.add_argument("dest", type=pathlib.Path, help="Destination directory/information.")
    return parser


def shelephant_cp(args: list[str], paths: list[str] = None):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    :param paths: Paths to copy (if not given, all files in source are copied).
    """

    parser = _shelephant_cp_parser()
    args = parser.parse_args(args)
    args.mode = args.mode.split(",")
    assert args.source.is_file(), "Source must be a file."
    assert not args.force if args.dry_run else True, "Cannot use --force with --dry-run."
    assert shutil.which("rsync") is not None or "rsync" not in args.mode, "rsync not available."
    assert "basic" not in args.mode if "rsync" in args.mode else True, "Use 'basic' or 'rsync'."

    if args.dest.is_file():
        dest = dataset.Location.from_yaml(args.dest)
    else:
        dest = dataset.Location(root=args.dest, ssh=args.ssh)

    source = dataset.Location.from_yaml(args.source)
    files = source.files(info=False)
    if paths is not None:
        files = np.intersect1d(files, paths).tolist()
    equal = []
    sourcepath = source.hostpath
    destpath = dest.hostpath

    if source.ssh is not None or dest.ssh is not None:
        assert "rsync" in args.mode, "'rsync' required for ssh."

    if "sha256" in args.mode:
        equal = source.diff(dest)["=="]
        equal = np.intersect1d(equal, files).tolist()
        files = np.setdiff1d(files, equal, assume_unique=True).tolist()  # based on sha256

    if "rsync" in args.mode:
        status = rsync.diff(sourcepath, destpath, files)
        eq = status.pop("==", [])
        [files.remove(file) for file in eq]  # based on rsync criteria
        equal += eq
    elif "basic" in args.mode:
        status = local.diff(sourcepath, destpath, files)

    assert status.pop("<-", []) == [], "Cannot copy from destination to source."

    if len(files) == 0:
        print("Nothing to copy" if len(equal) == 0 else "All files equal")
        return []

    if not args.force:
        status["=="] = equal
        output.copyplan(status, colors=args.colors)
        if args.dry_run:
            return []
        if not click.confirm("Proceed?"):
            raise OSError("Cancelled")

    if "rsync" in args.mode:
        rsync.copy(sourcepath, destpath, files, progress=not args.quiet)
    else:
        local.copy(sourcepath, destpath, files, progress=not args.quiet)

    return files


def _shelephant_mv_parser():
    """
    Return parser for :py:func:`shelephant_mv`.
    """

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    desc = textwrap.dedent(
        """
        Move files listed in a (field of a) YAML-file.
        These filenames are assumed either relative to the YAML-file or absolute.

        Usage::

            shelephant_mv <sourceinfo.yaml> <dest_dirname>
        """
    )

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("--colors", type=str, default="dark", help="Color scheme [none, dark].")
    parser.add_argument("-f", "--force", action="store_true", help="Overwrite without prompt.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("-n", "--dry-run", action="store_true", help="Print copy-plan and exit.")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("source", type=pathlib.Path, help="Source information.")
    parser.add_argument("dest", type=pathlib.Path, help="Destination directory.")
    return parser


def shelephant_mv(args: list[str], paths: list[str] = None):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    :param paths: Paths to move (if not given, all files in source are moved).
    """

    parser = _shelephant_mv_parser()
    args = parser.parse_args(args)
    assert args.source.is_file(), "Source must be a file."
    assert args.dest.is_dir(), "Destination must be a directory."
    assert not args.force if args.dry_run else True, "Cannot use --force with --dry-run."

    source = dataset.Location.from_yaml(args.source)
    assert source.ssh is None, "Cannot move from remote."
    files = source.files(info=False)
    if paths is not None:
        files = np.intersect1d(files, paths).tolist()
    sourcepath = source._absroot
    destpath = args.dest
    status = local.diff(sourcepath, destpath, files)
    assert status.pop("<-", []) == [], "Cannot move from destination to source"

    if len(files) == 0:
        print("Nothing to move")
        return

    if not args.force:
        output.copyplan(status, colors=args.colors)
        if args.dry_run:
            return
        if not click.confirm("Proceed?"):
            raise OSError("Cancelled")

    local.move(sourcepath, destpath, files, progress=not args.quiet)


def _shelephant_rm_parser():
    """
    Return parser for :py:func:`shelephant_rm`.
    """

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    desc = textwrap.dedent(
        """
        Remove files listed in a (field of a) YAML-file.
        These filenames are assumed either relative to the YAML-file or absolute.

        Usage::

            shelephant_rm <sourceinfo.yaml>
        """
    )

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("-f", "--force", action="store_true", help="Remove without prompt.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("-n", "--dry-run", action="store_true", help="Print copy-plan and exit.")
    parser.add_argument("--verbose", action="store_true", help="Print commands (only SSH remote).")
    parser.add_argument("-v", "--version", action="version", version=version)
    parser.add_argument("source", type=pathlib.Path, help="Source information.")
    return parser


def shelephant_rm(args: list[str], paths: list[str] = None):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    :param paths: Paths to remove (if not given, all files in source are removed).
    """

    parser = _shelephant_rm_parser()
    args = parser.parse_args(args)
    assert args.source.is_file(), "Source must be a file."
    assert not args.force if args.dry_run else True, "Cannot use --force with --dry-run."

    source = dataset.Location.from_yaml(args.source)
    files = source.files(info=False)
    if paths is not None:
        files = np.intersect1d(files, paths).tolist()
    sourcepath = source._absroot

    if len(files) == 0:
        print("Nothing to remove")
        return

    if not args.force:
        for file in files:
            print(f"rm {file:s}")
        if args.dry_run:
            return
        if not click.confirm("Proceed?"):
            raise OSError("Cancelled")

    if source.ssh is None:
        return local.remove(sourcepath, files, progress=not args.quiet)

    with ssh.tempdir(source.ssh) as remote, search.tempdir():
        files = [str(sourcepath / i) for i in files]
        pathlib.Path("remove.txt").write_text("\n".join(files))
        shutil.copy(pathlib.Path(__file__).parent / "_remove.py", "script.py")
        hostpath = f'{source.ssh:s}:"{str(remote):s}"'
        scp.copy(".", hostpath, ["script.py", "remove.txt"], progress=False, verbose=args.verbose)
        exec_cmd(
            f'ssh {source.ssh:s} "cd {str(remote)} && {source.python} script.py"',
            verbose=args.verbose,
        )


def _shelephant_hostinfo_parser():
    """
    Return parser for :py:func:`shelephant_hostinfo`.
    """

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    desc = textwrap.dedent(
        """
        Collect information about a remote directory (on a remote SSH host).
        This information is stored in a YAML-file (default: ``shelephant_hostinfo.yaml``)::

            root: <path>      # relative to the YAML-file, or absolute
            ssh: <user@host>  # (optional) remote SSH host
            dump: <dump>      # (optional, excludes "search") yaml-file to read list of files from
            search:           # (optional, excludes "dump") search information, must be set by hand
                - ...
            files:            # (optional) list of files (from "search" / "dump", or set by hand)
                - ...

        Usage:

        1.  Create *hostinfo*::

                # set "root"
                shelephant_hostinfo <path>

                # set "root" and "ssh"
                shelephant_hostinfo <path> --ssh <user@host>

                # set "root" (and "ssh") and "dump", and update "files"
                shelephant_hostinfo <path> --dump [shelephant_dump.yaml]
                shelephant_hostinfo <path> --dump [shelephant_dump.yaml] --ssh <user@host>

        2.  Update *hostinfo* (update "files")::

                # update "files" based on "dump" or "search"
                shelephant_hostinfo --update [shelephant_hostinfo.yaml]
        """
    )

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument(
        "-o", "--output", type=pathlib.Path, default=f_hostinfo, help="Output YAML-file."
    )
    parser.add_argument(
        "-d",
        "--dump",
        type=pathlib.Path,
        default=None,
        nargs="?",
        const=f_dump,
        help="YAML-file containing a list of files.",
    )
    parser.add_argument("--ssh", type=str, help="Remote SSH host (e.g. user@host).")
    parser.add_argument(
        "-u", "--update", action="store_true", help='Update "files" based on "dump" or "search".'
    )
    parser.add_argument("-i", "--info", action="store_true", help="Add information (sha256, size).")
    parser.add_argument("-f", "--force", action="store_true", help="Force overwrite output.")
    parser.add_argument("--verbose", action="store_true", help="Print commands (only SSH remote).")
    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("path", type=pathlib.Path, help="Path to remote directory.")
    return parser


def shelephant_hostinfo(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _shelephant_hostinfo_parser()
    args = parser.parse_args(args)

    if args.update:
        assert str(args.output) == f_hostinfo, "No customisation allowed in --update mode."
        assert args.dump is None, "No customisation allowed in --update mode."
        assert args.ssh is None, "No customisation allowed in --update mode."
        loc = dataset.Location.from_yaml(args.path)
        args.output = args.path
        args.force = True
    else:
        loc = dataset.Location(root=args.path, ssh=args.ssh)
        if args.dump:
            loc.dump = args.dump

    loc.read(verbose=args.verbose)
    if args.info:
        loc.getinfo(verbose=args.verbose)
    loc.to_yaml(args.output, force=args.force)


def _shelephant_diff_parser():
    """
    Return parser for :py:func:`shelephant_diff`.
    """

    desc = textwrap.dedent(
        """
        Compare local and remote files and list differences.

        Usage::
            shelephant_diff <sourceinfo.yaml> <dest_dirname>
            shelephant_diff <sourceinfo.yaml> <destinfo.yaml>
            shelephant_diff <sourceinfo.yaml> <destinfo.yaml> --filter "->"
            shelephant_diff <sourceinfo.yaml> <destinfo.yaml> --filter "?=, !="
            shelephant_diff <sourceinfo.yaml> <destinfo.yaml> -o <diff.yaml>

        Note that if filter contains only one operation the output YAML-file will be a list.
        """
    )

    desc = ""

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("--ssh", type=str, help="SSH destination (e.g. user@host).")
    parser.add_argument("--colors", type=str, default="dark", help="Color scheme [none, dark].")
    parser.add_argument(
        "--mode", type=str, help="Use 'sha256', 'rsync', or 'basic'.", default="sha256"
    )
    parser.add_argument("--sort", type=str, help="Sort printed table by column.")
    parser.add_argument("--table", type=str, default="SINGLE_BORDER", help="Select print style.")
    parser.add_argument("--filter", type=str, help="Filter to direction (separated by ',').")
    parser.add_argument("-o", "--output", type=pathlib.Path, help="Dump as YAML file.")
    parser.add_argument("-f", "--force", action="store_true", help="Force overwrite output.")
    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("source", type=pathlib.Path, help="Source information.")
    parser.add_argument("dest", type=pathlib.Path, help="Destination directory/information.")
    return parser


def shelephant_diff(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _shelephant_diff_parser()
    args = parser.parse_args(args)
    args.mode = args.mode.split(",")
    assert args.source.is_file(), "Source must be a file."
    assert len(args.mode) == 1, "Only one mode allowed."
    assert shutil.which("rsync") is not None or "rsync" not in args.mode, "rsync not available."

    source = dataset.Location.from_yaml(args.source)
    files = source.files(info=False)

    if args.dest.is_file():
        dest = dataset.Location.from_yaml(args.dest)
    else:
        dest = dataset.Location(root=args.dest, ssh=args.ssh)

    if "sha256" in args.mode:
        status = source.diff(dest)
    elif "rsync" in args.mode:
        left = source.diff(dest)["<-"]
        [files.remove(file) for file in left]
        status = rsync.diff(source.hostpath, dest.hostpath, files)
        status["<-"] = left
    elif "basic" in args.mode:
        assert source.ssh is None and dest.ssh is None, "Use 'rsync' or 'sha256' mode with SSH."
        status = local.diff(source.hostpath, dest.hostpath, files)
    else:
        raise ValueError(f"Unknown mode '{args.mode}'.")

    if args.filter:
        keys = [key.strip() for key in args.filter.split(",")]
        keys = [key for key in keys if key in status]
        status = {key: status[key] for key in keys}

    for key in list(status.keys()):
        if len(status[key]) == 0:
            del status[key]

    if args.output:
        if len(status) == 1:
            status = status[list(status.keys())[0]]
        yaml.dump(args.output, status, force=args.force)
        return

    out = prettytable.PrettyTable()
    if args.table == "PLAIN_COLUMNS":
        out.set_style(prettytable.PLAIN_COLUMNS)
    elif args.table == "SINGLE_BORDER":
        out.set_style(prettytable.SINGLE_BORDER)
    out.field_names = ["source", "sync", "dest"]
    out.align["source"] = "l"
    out.align["sync"] = "c"
    out.align["dest"] = "l"

    left = status.pop("->", [])
    right = status.pop("<-", [])
    equal = status.pop("==", [])

    for key in status:
        for item in status[key]:
            out.add_row([item, key, item])

    for item in left:
        out.add_row([item, "->", ""])

    for item in right:
        out.add_row(["", "<-", item])

    for item in equal:
        out.add_row([item, "==", item])

    if args.sort is None:
        print(out.get_string())
    else:
        print(out.get_string(sortby=args.sort))


def _shelephant_main_parser():
    """
    Return parser for :py:func:`shelephant`.
    """

    desc = textwrap.dedent(
        """
        Available commands:

        =========== ======================================================================
        command     description
        =========== ======================================================================
        init        Initialize a new dataset.
        add         Add storage location to dataset.
        remove      Remove storage location from dataset.
        update      Update dataset.
        status      Show status of files.
        info        Show global information about dataset.
        lock        Lock as storage location.
        cp          Copy files from one location to another.
        mv          Move files from one location to another (both local).
        rm          Remove files from one location.
        pwd         Print equivalent directory in the storage location.
        git         Run git command on the database directory (``.shelephant``).
        gitignore   Add all symbolic links at ``.shelephant`` to dataset's ``.gitignore``.
        =========== ======================================================================
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    choices = [
        "init",
        "add",
        "remove",
        "update",
        "status",
        "info",
        "lock",
        "cp",
        "mv",
        "rm",
        "pwd",
        "git",
        "gitignore",
    ]
    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("command", type=str, choices=choices, help="Command to run.")
    return parser


def _shelephant_main():
    assert len(sys.argv) >= 2, "No command given."
    parser = _shelephant_main_parser()
    args = parser.parse_args([sys.argv[1]])

    if args.command == "init":
        dataset.init(sys.argv[2:])
    elif args.command == "add":
        dataset.add(sys.argv[2:])
    elif args.command == "remove":
        dataset.remove(sys.argv[2:])
    elif args.command == "update":
        dataset.update(sys.argv[2:])
    elif args.command == "status":
        dataset.status(sys.argv[2:])
    elif args.command == "info":
        dataset.info(sys.argv[2:])
    elif args.command == "cp":
        dataset.cp(sys.argv[2:])
    elif args.command == "mv":
        dataset.mv(sys.argv[2:])
    elif args.command == "rm":
        dataset.rm(sys.argv[2:])
    elif args.command == "pwd":
        dataset.pwd(sys.argv[2:])
    elif args.command == "lock":
        dataset.lock(sys.argv[2:])
    elif args.command == "git":
        dataset.git(sys.argv[2:])
    elif args.command == "gitignore":
        dataset.gitignore(sys.argv[2:])
    else:
        raise ValueError(f"Unknown command '{sys.argv[1]}'.")


def _shelephant_parse_main():
    shelephant_parse(sys.argv[1:])


def _shelephant_cp_main():
    shelephant_cp(sys.argv[1:])


def _shelephant_mv_main():
    shelephant_mv(sys.argv[1:])


def _shelephant_rm_main():
    shelephant_rm(sys.argv[1:])


def _shelephant_dump_main():
    shelephant_dump(sys.argv[1:])


def _shelephant_hostinfo_main():
    shelephant_hostinfo(sys.argv[1:])


def _shelephant_diff_main():
    shelephant_diff(sys.argv[1:])



from . import cli
from . import compute_hash
from . import convert
from . import dataset
from . import local
from . import output
from . import path
from . import rsync
from . import scp
from . import search
from . import ssh
from . import yaml
from ._version import version



import os
import sys

sys.path.insert(0, os.path.abspath(".."))

project = "shelephant"
copyright = "2021, Tom de Geus"
author = "Tom de Geus"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.todo",
    "sphinxarg.ext",
]

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

html_theme = "furo"



import io
import os
import shlex
import shutil
import subprocess
import sys


def _theme(name: str = None) -> dict:
    r"""
    Return dictionary of colors.

    .. code-block:: python

        {
            'new' : '...',
            'overwrite' : '...',
            'skip' : '...',
            'bright' : '...',
        }

    :param name: Select color-theme [dark, none].
    :return: Dictionary of colors.
    """

    if name == "dark":
        return {
            "new": "1;32",
            "overwrite": "1;31",
            "skip": "1;30",
            "bright": "1;37",
        }

    return {
        "new": "",
        "overwrite": "",
        "skip": "",
        "bright": "",
    }


def _format(text: str, width: int = None, align: str = "<", color: str = None) -> str:
    r"""
    Format with color and alignment.

    :param text: The plain text.
    :param width: Print width.
    :param color: Print color, e.g. "1;32" for bold green.
    :param align: Print alignment.
    :return: Formatted string.
    """

    if width and color:
        fmt = "\x1b[{color:s}m{{0:{align:s}{width:d}.{width:d}s}}\x1b[0m".format(
            width=width, align=align, color=color
        )
    elif width:
        fmt = "{{0:{align:s}{width:d}.{width:d}s}}".format(width=width, align=align)
    elif color:
        fmt = f"\x1b[{color:s}m{{0:{align:s}s}}\x1b[0m"
    else:
        fmt = f"{{0:{align:s}s}}"

    return fmt.format(text)


def _page(text: str):
    """
    Display text in a terminal pager.
    Respects the PAGER environment variable if set.

    :param text: Text to display.
    """
    pager_cmd = shlex.split(os.environ.get("PAGER") or "less -r")
    subprocess.run(pager_cmd, input=text.encode("utf-8"))


def copyplan(
    status: dict[list[str]],
    colors: str = "none",
    display: bool = True,
    max_align: int = 80,
) -> str:
    """
    Print copy plan.

    :param status:
        Dictionary of copy status. E.g.::

            {
                '->' : ['file1', 'file2'],
                '!=' : ['file3'],
                '==' : ['file4'],
            }

    :param colors: Color theme name, see :py:func:`theme`.
    :param display: Display output (``False``: return as string).
    :param max_align: Maximum width of the first column.
    :return: Output string (if ``display=False``).
    """
    color = _theme(colors.lower())
    sio = io.StringIO()

    assert status.pop("<-", []) == [], "Cannot copy from destination to source"
    skip = status.pop("==", [])
    right = status.pop("->", [])
    overwrite = []
    for key in list(status.keys()):
        overwrite += status.pop(key, [])

    if len(overwrite) + len(right) + len(skip) == 0:
        return

    width = max(len(file) for file in overwrite + right + skip)
    width = min(width, max_align)

    for file in overwrite:
        print(
            "{:s} {:s} {:s}".format(
                _format(file, width=width, color=color["bright"]),
                _format("=>", color=color["bright"]),
                _format(file, color=color["overwrite"]),
            ),
            file=sio,
        )

    for file in right:
        print(
            "{:s} {:s} {:s}".format(
                _format(file, width=width, color=color["bright"]),
                _format("->", color=color["bright"]),
                _format(file, color=color["new"]),
            ),
            file=sio,
        )

    for file in skip:
        print(
            "{:s} {:s} {:s}".format(
                _format(file, width=width, color=color["skip"]),
                _format("==", color=color["skip"]),
                _format(file, color=color["skip"]),
            ),
            file=sio,
        )

    if not display:
        return sio.getvalue()

    output = sio.getvalue()
    if sys.stdout.isatty():
        nlines = len(output.splitlines())
        _, term_lines = shutil.get_terminal_size()
        if nlines > term_lines:
            return _page(output)

    print(output)



import hashlib
import os
import pathlib
import sys

try:
    from tqdm import tqdm
except ImportError:

    def tqdm(iterator, *args, **kwargs):
        return iterator


def compute_sha256(
    files: list[pathlib.Path], sha256: bool = True, progress: bool = True
) -> tuple[list[str], list[int]]:
    """
    Get the sha256 hash and size of a list of files.

    :param files: A list of files.
    :param sha256: Calculate the sha256 hash.
    :param progress: Show a progress bar.
    :return: A tuple of lists of (size, mtime, sha256).
    """

    ret_hash = []
    ret_size = []
    ret_mtime = []

    if not sha256:
        for filename in tqdm(files, disable=not progress):
            if not os.path.exists(filename):
                ret_size.append(-1)
                ret_mtime.append(-1)
                ret_hash.append("")
                continue
            ret_size.append(os.path.getsize(filename))
            ret_mtime.append(os.path.getmtime(filename))

    elif sys.version_info >= (3, 11):
        for filename in tqdm(files, disable=not progress):
            if not os.path.exists(filename):
                ret_size.append(-1)
                ret_mtime.append(-1)
                ret_hash.append("")
                continue
            ret_size.append(os.path.getsize(filename))
            ret_mtime.append(os.path.getmtime(filename))
            with open(filename, "rb", buffering=0) as f:
                ret_hash.append(hashlib.file_digest(f, "sha256").hexdigest())

    else:
        for filename in tqdm(files, disable=not progress):
            if not os.path.exists(filename):
                ret_size.append(-1)
                ret_mtime.append(-1)
                ret_hash.append("")
                continue
            ret_size.append(os.path.getsize(filename))
            ret_mtime.append(os.path.getmtime(filename))
            h = hashlib.sha256()
            b = bytearray(128 * 1024)
            mv = memoryview(b)
            with open(filename, "rb", buffering=0) as f:
                while n := f.readinto(mv):
                    h.update(mv[:n])
            ret_hash.append(h.hexdigest())

    return ret_size, ret_mtime, ret_hash


if __name__ == "__main__":
    sha256 = pathlib.Path("sha256.txt").exists()
    size, mtime, csum = compute_sha256(
        pathlib.Path("files.txt").read_text().splitlines(), sha256=sha256
    )
    pathlib.Path("size.txt").write_text("\n".join(map(str, size)))
    pathlib.Path("mtime.txt").write_text("\n".join(map(str, mtime)))
    pathlib.Path("sha256.txt").write_text("\n".join(csum))



import pathlib
import shutil
import unittest

import shelephant
from shelephant._tests import create_dummy_files
from shelephant.search import cwd
from shelephant.search import tempdir

has_rsync = shutil.which("rsync") is not None


class Test_local(unittest.TestCase):
    def test_diff(self):
        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                data = shelephant.local.diff(".", "../dest", files)

            check = {
                "?=": ["foo.txt", "bar.txt"],
                "->": ["more.txt", "even_more.txt"],
                "<-": [],
            }
            self.assertEqual(data, check)

    def test_copy(self):
        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant.local.copy(".", "../dest", files, progress=False)

            with cwd("dest"):
                data = shelephant.dataset.Location(root=".", files=files).getinfo()

            self.assertTrue(check == data)


class Test_scp(unittest.TestCase):
    def test_copy(self):
        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant.scp.copy(".", "../dest", files, progress=False)

            with cwd("dest"):
                data = shelephant.dataset.Location(root=".", files=files).getinfo()

            self.assertTrue(check == data)


class Test_rsync(unittest.TestCase):
    def test_diff(self):
        if not has_rsync:
            self.skipTest("rsync not found")

        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                data = shelephant.rsync.diff(".", "../dest", files)

            check = {
                "==": ["foo.txt"],
                "!=": ["bar.txt"],
                "->": ["more.txt", "even_more.txt"],
            }
            self.assertEqual(data, check)

    def test_copy(self):
        if not has_rsync:
            self.skipTest("rsync not found")

        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant.rsync.copy(".", "../dest", files, progress=False)

            with cwd("dest"):
                data = shelephant.dataset.Location(root=".", files=files).getinfo()

            self.assertTrue(check == data)


if __name__ == "__main__":
    unittest.main()



"""
Not part of public API.
"""
import os
import pathlib

import numpy as np

from . import dataset


def create_dummy_files(filenames: list[str], keep: list = None) -> dataset.Location:
    """
    Create dummy files in the current directory.

    :param filenames: List of filenames.
    :param keep: Select a subset of available dummy files.
        For example ``keep=[0, -1]`` or ``keep=slice(0, None, 2)``.

    :return: dataset.Location
    """

    content = {
        "foo": "2c26b46b68ffc68ff99b453c1d30413413422d706483bfa0f98a5e886266e7ae",
        "bar": "fcde2b2edba56bf408601fb721fe9b5c338d10ee429ea04fae5511b68fbf8fb9",
        "a": "ca978112ca1bbdcafac231b39a23dc4da786eff8147c4e72b9807785afee48bb",
        "b": "3e23e8160039594a33894f6564e1b1348bbd7a0088d42c4acb73eeaed59c009d",
        "c": "2e7d2c03a9507ae265ecf5b5356885a53393a2029d241394997265a1a25aefc6",
        "d": "18ac3e7343f016890c510e93f935261169d9e3f565436429830faf0934f4f8e4",
        "e": "3f79bb7b435b05321651daefd374cdc681dc06faa65e374e38337b88ca046dea",
        "f": "252f10c83610ebca1a059c0bae8255eba2f95be4d1d7bcfa89d7248a82d9f111",
        "g": "cd0aa9856147b6c5b4ff2b7dfee5da20aa38253099ef1b4a64aced233c9afe29",
    }

    if keep is not None:
        keys = list(content.keys())
        keep = np.arange(len(keys))[keep]
        content = {keys[i]: content[keys[i]] for i in keep}

    assert len(filenames) <= len(content)

    ret = {}
    for file, (content, sha) in zip(filenames, content.items()):
        pathlib.Path(file).write_text(content)
        ret[file] = {
            "sha256": sha,
            "size": os.path.getsize(file),
            "mtime": os.path.getmtime(file),
        }

    return dataset.Location(root=".", files=ret)



import collections.abc
import functools
import operator


def _flatten_detail(data):
    r"""
    Detail of :py:fun:`Flatten`.
    See https://stackoverflow.com/a/17485785/2646505
    """

    for item in data:
        if isinstance(item, collections.abc.Iterable) and not isinstance(item, str):
            yield from flatten(item)
        else:
            yield item


def flatten(data: list[list]) -> list:
    """
    Flatten a nested list to a one dimensional list.

    :param data: A nested list.
    :return: A one dimensional list.
    """
    return list(_flatten_detail(data))


def _squash_detail(data, parent_key="", sep="_"):
    r"""
    Detail of :py:fun:`squash`.
    See https://stackoverflow.com/a/6027615/2646505
    """

    items = []

    for k, v in data.items():
        new_key = parent_key + sep + k if parent_key else k

        if isinstance(v, collections.abc.MutableMapping):
            items.extend(_squash_detail(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))

    return dict(items)


def squash(data: dict[list]) -> list:
    """
    Squash a dictionary to a single list.
    For example::

        >>> squash({"foo": [1, 2], "bar": {"foo": [3, 4], "bar": 5}})
        [1, 2, 3, 4, 5]

    :param data: A nested dictionary.
    :return: A one dimensional list.
    """

    return flatten(list(_squash_detail(data).values()))


def split_key(key: str) -> list[str]:
    """
    Split a key separated by "/" in a list.

    :param key: A key.
    :return: A list of key components.
    """

    if type(key) == list:
        return key

    if type(key) == str:
        return key.split("/")

    raise OSError(f"'{key}' cannot be split")


def get(data: dict[dict], key: str | list[str]) -> dict | list | str | int | float:
    r"""
    Get an item from a nested dictionary.

    :param data: A nested dictionary.
    :param key:
        The item to read. E.g.
        *   ``[]`` for a YAML file containing only a list.
        *   ``['foo']`` for a plain YAML file.
        *   ``['key', 'to', foo']`` for a YAML file with nested items.

        An item specified as ``str`` separated by "/" is also accepted.

    :return:
        The read item.
    """

    assert type(data) == dict
    key = split_key(key)

    if len(key) > 0:
        try:
            return functools.reduce(operator.getitem, key, data)
        except KeyError:
            raise OSError('"{:s}" not found'.format("/".join(key)))

    return data



import argparse
import json
import os
import pathlib
import re
import shutil
import textwrap
from copy import deepcopy

import numpy as np
import prettytable
import tqdm
from platformdirs import user_cache_dir

from . import cli
from . import compute_hash
from . import rsync
from . import scp
from . import search
from . import ssh
from . import yaml
from ._version import version
from .external import exec_cmd

if shutil.which("rsync") is not None:
    _copyfunc = rsync.copy
else:
    _copyfunc = scp.copy


def _force_absolute_path(root: pathlib.Path, path: pathlib.Path) -> pathlib.Path:
    """
    Force a path to be absolute.

    :param root: The root directory.
    :param path: The path that may be absolute or relative.
    :return: The absolute ``path``.
    """
    if path.is_absolute():
        return path
    return pathlib.Path(os.path.normpath(root.absolute() / path))


class Location:
    """
    Location information.

    Attributes:

    *   :py:attr:`Location.root`: The root directory.
    *   :py:attr:`Location.ssh` (optional): ``[user@]host``
    *   :py:attr:`Location.prefix` (optional): Prefix to add to all paths.
    *   :py:attr:`Location.python` (optional): The python executable on the ``ssh`` host.
    *   :py:attr:`Location.dump` (optional): Location of "dump" file -- file with list of files.
    *   :py:attr:`Location.search` (optional):
        Commands to search for files, see :py:func:`shelephant.search.search`.
    *   :py:func:`Location.files`: List of files (with properties).

    Initialize:

    *   Read from yaml file::

            location = Location.from_yaml("location.yaml")

    *   Create from scratch::

            location = Location(root="~/data", ...)
    """

    def __init__(
        self,
        root: str | pathlib.Path,
        ssh: str = None,
        mount: pathlib.Path = None,
        prefix: pathlib.Path = None,
        files: list[str] = [],
    ) -> None:
        """
        :param root: The root directory (may be relative, unless on remote SSH host).
        :param ssh: ``[user@]host``.
        :param mount: Mount location for SSH host.
        :param prefix: Prefix to add to all paths.
        :param files: List of files.
        """
        self.root = pathlib.Path(root)
        self._mount = mount is not None
        self._absroot = self.root.absolute() if not mount else mount.absolute()
        self.prefix = pathlib.Path(prefix) if prefix is not None else None
        self.ssh = ssh
        self.python = "python3"
        self.dump = None
        self.search = None

        if ssh is not None:
            assert self.root.is_absolute(), "root must be absolute path when using ssh"

        if type(files) == list:
            self._files = np.array(files)
            self._clear_all_info()
        elif type(files) == str:
            self._files = np.array([files])
            self._clear_all_info()
        elif type(files) == dict:
            self._files = np.array(list(files.keys()))
            self._clear_all_info()
            for i, file in enumerate(self._files):
                self._has_info[i] = "sha256" in files[file]
                if self._has_info[i]:
                    self._sha256[i] = files[file]["sha256"]
                    self._size[i] = files[file]["size"]
                    self._mtime[i] = files[file]["mtime"]
        else:
            raise TypeError(f"Unknown type of files: {type(files)}")

        assert np.unique(self._files).size == self._files.size, "duplicate filenames"

    def _slice(self, keep):
        """
        Slice list of files and info.

        :param keep: Slice.
        """
        self._files = self._files[keep]
        self._has_info = self._has_info[keep]
        self._sha256 = self._sha256[keep]
        self._size = self._size[keep]
        self._mtime = self._mtime[keep]
        return self

    def _clear_all_info(self):
        """
        Clear all info.
        """
        self._has_info = np.zeros(self._files.size, dtype=bool)
        self._sha256 = np.empty(self._files.size, dtype="U64")
        self._size = np.zeros(self._files.size, dtype=np.int64)
        self._mtime = np.empty(self._files.size, dtype=np.float64)
        return self

    def _append(self, files: list[str]):
        """
        Extend list of files.

        .. warning::

            ``files`` is assumed to be unique and not part of ``self._files``.
            There is no check on this.

        :param files: List of files.
        """
        files = np.array(files)
        n = files.size
        if n == 0:
            return self
        self._files = np.hstack((self._files, files))
        self._has_info = np.hstack((self._has_info, np.zeros(n, dtype=self._has_info.dtype)))
        self._sha256 = np.hstack((self._sha256, np.empty(n, dtype=self._sha256.dtype)))
        self._size = np.hstack((self._size, np.empty(n, dtype=self._size.dtype)))
        self._mtime = np.hstack((self._mtime, np.empty(n, dtype=self._mtime.dtype)))
        if self.search is not None:
            return self.sort()
        return self

    def _prune(self, files: list[str]):
        """
        Overwrite database with list of files.
        If files are already in the database, the sha256/size/mtime is kept.
        For new files the sha256/size/mtime is set to None.
        (Files that were in the database but that are not in ``files`` are removed.)

        :param files: List of files (unique, no assertion).
        """
        files = np.array(files)

        if self._files.size == files.size:
            if np.all(np.equal(self._files, files)):
                return self

        _, i, j = np.intersect1d(self._files, files, return_indices=True, assume_unique=True)
        k = np.ones(files.size, dtype=bool)
        k[j] = False
        self._slice(i)
        self._append(files[k])
        return self

    def _overwrite_dataset_from_dict(self, files: list):
        """
        Read files from list of strings or dictionaries (as stored in a YAML file).
        This overwrites the current database (including all sha256/size/mtime), with information
        present in the input.

        :param files: List of files.
        """

        fs = []
        has_info = []
        sha256 = []
        size = []
        mtime = []

        for item in files:
            if type(item) == str:
                fs.append(item)
                has_info.append(False)
                sha256.append("0" * 64)
                size.append(0)
                mtime.append(0)
            else:
                fs.append(item["path"])
                has_info.append(True)
                sha256.append(item["sha256"])
                size.append(item["size"])
                mtime.append(item["mtime"])

        self._files = np.array(fs, dtype=object)
        self._has_info = np.array(has_info, dtype=bool)
        self._sha256 = np.array(sha256, dtype="U64")
        self._size = np.array(size, dtype=np.int64)
        self._mtime = np.array(mtime, dtype=np.float64)
        return self

    def __eq__(self, other):
        """
        Check if all files and information are equal (mtime is allowed to be different).

        :param other: Other location.
        :return: True if equal, False otherwise.
        """
        if self._files.size != other._files.size:
            return False

        a = np.argsort(self._files)
        b = np.argsort(other._files)

        return (
            np.all(np.equal(self._files[a], other._files[b]))
            and np.all(np.equal(self._sha256[a], other._sha256[b]))
            and np.all(np.equal(self._size[a], other._size[b]))
            and np.all(np.equal(self._has_info[a], other._has_info[b]))
        )

    def __iadd__(self, other):
        """
        Add files from other location.

        .. todo::

            In case of duplicate files: check that the checksums and sizes are the same.
        """
        assert self.root == other.root, "root must be equal"
        assert self.ssh == other.ssh, "ssh must be equal"
        self._files = np.hstack((self._files, other._files))
        self._has_info = np.hstack((self._has_info, other._has_info))
        self._sha256 = np.hstack((self._sha256, other._sha256))
        self._size = np.hstack((self._size, other._size))
        self._mtime = np.hstack((self._mtime, other._mtime))
        self.search = None
        self.dump = None
        return self._unique()

    def __add__(self, other):
        """
        Add files from two locations.
        """
        ret = deepcopy(self)
        ret += other
        return ret._unique()

    def copy_files(self, other):
        """
        Copy files (and size/mtime/sha256) from other location.

        :param other: Other location.
        """
        self._files = other._files
        self._has_info = other._has_info
        self._sha256 = other._sha256
        self._size = other._size
        self._mtime = other._mtime
        return self

    @classmethod
    def from_yaml(cls, path: str | pathlib.Path):
        """
        Read from yaml file.

        :param path: Path to yaml file.
        :return: Location.
        """

        path = pathlib.Path(path)
        data = yaml.read(path)
        if type(data) == list:
            data = {"files": data}

        ret = cls(
            root=data.get("root", path.parent),
            ssh=data.get("ssh", None),
            prefix=data.get("prefix", None),
        )
        ret._mount = "mount" in data
        assert not ret._mount or ret.ssh is not None, "ssh must be set when using mount"
        ret._absroot = data.get("mount", _force_absolute_path(path.parent, ret.root))
        ret.dump = data.get("dump", None)
        ret.search = data.get("search", None)
        ret._overwrite_dataset_from_dict(data.get("files", []))
        if ret.search is not None:
            ret.sort()
        return ret

    def to_yaml(self, path: str | pathlib.Path, force: bool = False):
        """
        Write to yaml file.

        :param path: Path to yaml file.
        :param force: Do not prompt to overwrite file.
        """
        yaml.dump(path, self.asdict(), force=force)

    def overwrite_yaml(self, path: str | pathlib.Path):
        """
        Overwrite yaml file.
        This function only changes the file if the content has indeed changed.

        :param path: Path to yaml file.
        """
        yaml.overwrite(path, self.asdict())

    def asdict(self) -> dict:
        """
        Return as dictionary.

        :return: Dictionary.
        """

        ret = {"root": str(self.root)}

        if self.ssh is not None:
            ret["ssh"] = self.ssh

        if self.dump is not None:
            ret["dump"] = str(self.dump)

        if self.search is not None:
            ret["search"] = self.search

        if self.prefix is not None:
            ret["prefix"] = str(self.prefix)

        if len(self._files) > 0:
            ret["files"] = self.files(info=True)

        return ret

    def files(self, info: bool = True) -> list:
        """
        Return as list of files.
        Items without sha256 and size are returned as ``str``,
        items with sha256 and size are returned as ``dict``.

        :param info: Return sha256 and size (if available).
        :return: List of files.
        """

        if not info:
            return self._files.tolist()

        if not np.any(self._has_info):
            return self._files.tolist()

        ret = []

        for i, file in enumerate(self._files):
            if self._has_info[i]:
                ret += [
                    {
                        "path": str(file),
                        "sha256": str(self._sha256[i]),
                        "size": int(self._size[i]),
                        "mtime": float(self._mtime[i]),
                    }
                ]
            else:
                ret += [str(file)]

        return ret

    @property
    def hostpath(self) -> str:
        """
        Return:

        -   ``root`` if ``ssh`` is not set.
        -   ``ssh:"root"`` if ``ssh`` is set.
        """
        if self.ssh is None:
            return str(self._absroot)

        return f'{self.ssh:s}:"{str(self.root):s}"'

    def sort(self, key: str = "files"):
        """
        Sort files.

        :param key: Key to sort by (files, size, sha256).
        """

        if key == "files":
            return self._slice(np.argsort(self._files))
        if key == "size":
            return self._slice(np.argsort(self._size))
        if key == "mtime":
            return self._slice(np.argsort(self._mtime))
        if key == "sha256":
            return self._slice(np.argsort(self._sha256))

        raise ValueError(f"Unknown key: {key}")

    def _unique(self):
        """
        Remove duplicate filename.

        .. todo::

            Check that the checksums and sizes are the same.

        .. todo::

            Keep sorted?
        """
        _, idx = np.unique(self._files, return_index=True)
        return self._slice(idx)

    def isavailable(self, mount: bool = False) -> bool:
        """
        Check if location is available.

        :param mount: Check if mount is available.
        :return: True if available.
        """
        if self.ssh is None or mount:
            return self._absroot.is_dir()
        return ssh.is_dir(self.ssh, self.root)

    def remove(self, paths: list[str]):
        """
        Remove files from list of files.

        :param paths: List of paths to remove.
        """

        _, i, _ = np.intersect1d(
            self._files, np.unique(list(map(str, paths))), return_indices=True, assume_unique=True
        )
        k = np.ones(self._files.size, dtype=bool)
        k[i] = False
        return self._slice(k)

    def read(self, verbose: bool = False):
        """
        Read files from location.

        -   If ``dump`` is set, read from dump file.
            This overwrites the database
            (sha256/size/mtime will only be available if they are in the YAML file).

        -   If ``search`` is set, search for files.
            This preserves sha256/size/mtime if paths are already in the database
            (there is no check that they are still accurate).

        :param verbose: Print progress (only relevant if ``ssh`` is set).
        """
        if self.dump is None and self.search is None:
            return self

        assert not (self.dump is not None and self.search is not None)

        # overwrite dataset with content of YAML dump file, and quit
        if self.dump is not None:
            if self.ssh is None:
                return self._overwrite_dataset_from_dict(yaml.read(self._absroot / self.dump))

            with search.tempdir():
                scp.copy(self.hostpath, ".", [self.dump], progress=False)
                return self._overwrite_dataset_from_dict(yaml.read(self.dump))

        # search locally for files (the sha256/size/mtime of 'new' files is set to None)
        if self.ssh is None:
            return self._prune(
                sorted(list(map(str, search.search(*self.search, root=self._absroot))))
            )

        # search on SSH remote host for files (the sha256/size/mtime of 'new' files is set to None)
        cache_dir = ssh._shelephant_cachdir(self.ssh, self.python)
        with ssh._cachedir(self.ssh, cache_dir) as remote, search.tempdir():
            shutil.copy(pathlib.Path(__file__).parent / "search.py", "script.py")
            with open("settings.json", "w") as f:
                json.dump(self.search, f)

            host = f'{self.ssh:s}:"{str(remote):s}"'
            _copyfunc(".", host, ["script.py", "settings.json"], progress=False, verbose=verbose)
            exec_cmd(
                f'ssh {self.ssh:s} "cd {str(remote)} && {self.python} script.py {str(self.root)}"',
                verbose=verbose,
            )
            _copyfunc(host, ".", ["files.txt"], progress=False, verbose=verbose)
            return self._prune(sorted(pathlib.Path("files.txt").read_text().splitlines()))

    def has_info(self) -> bool:
        """
        Check if sha256/size/mtime is available for all files.

        :return: True if sha256/size/mtime is available for all files.
        """
        return np.all(self._has_info)

    def _getindex(self, paths: list[str]) -> np.array:
        """
        Get index of paths in dataset.

        :param paths: List of paths.
        :return: ``paths`` (sorted) and their indices in ``self._files``.
        """
        paths = np.sort(paths)
        sorter = np.argsort(self._files)
        index = sorter[np.searchsorted(self._files[sorter], paths)]
        assert np.all(self._files[index] == paths), "not all paths are in the dataset"
        return paths, index

    def _get_info(self, paths: list[pathlib.Path], sha256: bool, progress: bool, verbose: bool):
        """
        Get mtime/size/sha256 of a list of files.

        :param paths: List of paths to check.
        :param progress: Show progress bar (only relevant if ``ssh`` is not set).
        :param verbose: Show verbose output (only relevant if ``ssh`` is set).
        :return: size, mtime, sha256
        """

        if self.ssh is None:
            files = [self._absroot / f for f in paths]
            size, mtime, csum = compute_hash.compute_sha256(files, sha256=sha256, progress=progress)
            return (
                np.array(size, dtype=np.int64),
                np.array(mtime, dtype=np.float64),
                np.array(csum, dtype="U64"),
            )

        cache_dir = ssh._shelephant_cachdir(self.ssh, self.python)
        with ssh._cachedir(self.ssh, cache_dir) as remote, search.tempdir():
            files = [str(self.root / i) for i in paths]
            pathlib.Path("files.txt").write_text("\n".join(files))
            pathlib.Path("sha256.txt").write_text("")
            shutil.copy(pathlib.Path(__file__).parent / "compute_hash.py", "script.py")

            extra = ["sha256.txt"] if sha256 else []
            hostpath = f'{self.ssh:s}:"{str(remote):s}"'
            _copyfunc(
                ".", hostpath, extra + ["script.py", "files.txt"], progress=False, verbose=verbose
            )
            exec_cmd(
                f'ssh {self.ssh:s} "cd {str(remote)} && {self.python} script.py"', verbose=verbose
            )
            _copyfunc(
                hostpath, ".", extra + ["size.txt", "mtime.txt"], progress=False, verbose=verbose
            )
            size = np.array(
                list(map(int, pathlib.Path("size.txt").read_text().splitlines())),
                dtype=np.int64,
            )
            mtime = np.array(
                list(map(float, pathlib.Path("mtime.txt").read_text().splitlines())),
                dtype=np.float64,
            )
            if sha256:
                csum = np.array(pathlib.Path("sha256.txt").read_text().splitlines(), dtype="U64")
            else:
                csum = []

        return size, mtime, csum

    def check_changes(
        self, paths: list[pathlib.Path] = None, progress: bool = False, verbose: bool = False
    ):
        """
        Remove sha256 from all files of which the size/mtime has changed.

        :param paths:
            List of paths to check.
            Paths to be relative to the root of the dataset.
            Default: all files in the database.

        :param progress: Show progress bar (only relevant if ``ssh`` is not set).
        :param verbose: Show verbose output (only relevant if ``ssh`` is set).
        """
        if paths is None:
            size, mtime, _ = self._get_info(self._files, False, progress, verbose)
            rm = np.logical_or(self._size != size, self._mtime != mtime)
            self._has_info[rm] = False
            self._size[~self._has_info] = size[~self._has_info]
            self._mtime[~self._has_info] = mtime[~self._has_info]
            return self

        paths, index = self._getindex(paths)
        size, mtime, _ = self._get_info(paths, False, progress, verbose)

        changed = np.logical_or(self._size[index] != size, self._mtime[index] != mtime)
        self._has_info[index[changed]] = False
        keep = ~self._has_info[index]
        self._size[index[keep]] = size[keep]
        self._mtime[index[keep]] = mtime[keep]

        removed = mtime < 0
        if np.any(removed):
            return self.remove(paths[removed])

        return self

    def remove_info(self, paths: list[pathlib.Path] = None):
        """
        Remove sha256/size/mtime for a list of files.

        :param paths: List of paths to remove.
        """
        paths, index = self._getindex(paths)
        self._has_info[index] = False

    def getinfo(
        self,
        paths: list[pathlib.Path] = None,
        max_size: int = None,
        progress: bool = False,
        verbose: bool = False,
    ):
        """
        Compute sha256/size/mtime of all files for which this information is not available.

        To compute the sha256/size/mtime only on a fraction of files set ``max_size``.
        This will stop the computation when the total size exceeds ``max_size``.
        You can then call this function recursively (with ``clean=False``) to flush you buffer.

        :param paths: List of paths to check. Have to be relative to the root of the dataset.
        :param max_size: Compute the sha256/size/mtime until the total size exceeds ``max_size``.
        :param progress: Show progress bar (only relevant if ``ssh`` is not set).
        :param verbose: Show verbose output (only relevant if ``ssh`` is set).
        """

        if paths is None:
            paths = self._files
            index = np.arange(self._files.size)
        else:
            paths, index = self._getindex(paths)

        index = index[np.argwhere(~self._has_info[index]).flatten()]
        index = index[np.argsort(self._size[index])]
        files = self._files[index]
        size = self._size[index]

        if index.size == 0:
            return self

        if max_size is not None:
            cum_size = np.cumsum(size)
            i = np.argmax(cum_size > max_size)
            if i > 0:
                index = index[:i]
                files = files[:i]

        size, mtime, csum = self._get_info(files, True, progress, verbose)
        self._has_info[index] = True
        self._sha256[index] = csum
        self._size[index] = size
        self._mtime[index] = mtime

        removed = mtime < 0
        if np.any(removed):
            return self.remove(paths[removed])

        return self

    def diff(self, other) -> dict:
        """
        Compare the database entries of two locations.

        .. warning::

            The information is taken from the database.
            There are no checks that that information if up to date.

        :param other: Other location.
        :return:
            Dictionary with differences::

                {
                    "==" : [ ... ], # in a and b, equal sha256
                    "?=" : [ ... ], # in a and b, unknown sha256
                    "!=" : [ ... ], # in a and b, different sha256
                    "->" : [ ... ], # in a not in b
                    "<-" : [ ... ], # in b not in a
                }
        """

        inboth = [str(file) for file in np.intersect1d(self._files, other._files)]
        ret = {
            "->": [str(file) for file in np.setdiff1d(self._files, other._files)],
            "<-": [str(file) for file in np.setdiff1d(other._files, self._files)],
            "==": [],
            "?=": [],
            "!=": [],
        }

        index_self = {file: i for i, file in enumerate(self._files)}
        index_other = {file: i for i, file in enumerate(other._files)}

        for file in inboth:
            ia = index_self[file]
            ib = index_other[file]

            if self._has_info[ia] and other._has_info[ib]:
                if self._sha256[ia] == other._sha256[ib]:
                    ret["=="].append(file)
                else:
                    ret["!="].append(file)
            else:
                ret["?="].append(file)

        return ret


# https://stackoverflow.com/a/68994012/2646505
def _search_upwards_dir(dirname: str) -> pathlib.Path:
    """
    Search in the current directory and all directories above it for a directory
    with a particular name.

    :param dirname: The name of the directory to look for.
    :return: The location of the first directory found or None, if none was found.
    """
    d = pathlib.Path.cwd()
    root = pathlib.Path(d.root)

    while d != root:
        attempt = d / dirname
        if attempt.is_dir():
            return attempt
        d = d.parent

    return None


def _init_parser():
    """
    Return parser for :py:func:`shelephant init`.
    """

    desc = textwrap.dedent(
        """
        Initialize a shelephant database by creating a directory ``.shelephant``
        with an 'empty' database. Use ``shelephant add`` to add storage locations.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("--version", action="version", version=version)
    return parser


def init(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _init_parser()
    args = parser.parse_args(args)
    sdir = pathlib.Path(".shelephant")
    assert not sdir.is_dir(), '".shelephant" directory already exists'
    sdir.mkdir()
    (sdir / "data").mkdir()
    (sdir / "storage").mkdir()
    (sdir / "unavailable").symlink_to("dead-link")
    (sdir / "symlinks.yaml").write_text("")
    yaml.dump(sdir / "storage.yaml", ["here"])
    yaml.dump(sdir / "storage" / "here.yaml", {"root": "../.."})


def _lock_parser():
    """
    Return parser for :py:func:`shelephant lock`.
    """

    desc = textwrap.dedent(
        """
        Lock as storage location.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)
    parser.add_argument("name", type=str, help="Name of the storage location.")
    parser.add_argument("--version", action="version", version=version)
    return parser


def lock(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _lock_parser()
    args = parser.parse_args(args)
    sdir = pathlib.Path(".shelephant")
    assert args.name in yaml.read(sdir / "storage.yaml"), f"storage location '{args.name}' unknown"
    (sdir / "lock.txt").write_text(args.name)


def _add_parser():
    """
    Return parser for :py:func:`shelephant add`.
    """

    desc = textwrap.dedent(
        """
        Add a storage location to the database.
        The database in ``.shelephant`` is updated as follows:

        -   The ``name`` is added to ``.shelephant/storage.yaml``.

        -   A file ``.shelephant/storage/<name>.yaml`` is created with the search settings
            and the present state of the storage location.

        -   A symlink ``.shelephant/data/<name>`` is created to the storage location.
            (if ``--ssh`` is given, the symlink points to a dead link).

        .. note::

            A special case is

            .. code-block:: bash

                shelephant add here --rglob '*.h5'

            which helps to investigate your database directory.
            Note that ``here`` is a reserved name and that you should not specify the root.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("name", type=str, help="Name of the storage location.")
    parser.add_argument("root", type=pathlib.Path, nargs="?", help="Path to the storage location.")
    parser.add_argument("--ssh", type=str, help="SSH host (e.g. user@host).")
    parser.add_argument("--mount", type=pathlib.Path, help="Optional mount location for SSH host.")
    parser.add_argument("--prefix", type=pathlib.Path, help="Add prefix to all files.")
    parser.add_argument("--rglob", type=str, help="Search pattern for ``Path(root).rglob(...)``.")
    parser.add_argument("--glob", type=str, help="Search pattern for ``Path(root).glob(...)``.")
    parser.add_argument("--exec", type=str, help="Command to run from ``root``.")
    parser.add_argument("--skip", type=str, action="append", help="Pattern to skip (Python regex).")
    parser.add_argument("--shallow", action="store_true", help="Do not compute checksums.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("--version", action="version", version=version)
    return parser


def add(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _add_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    assert not (sdir / "lock.txt").exists(), "cannot remove location from storage location"
    assert args.name != "all", "all is a reserved name"
    assert args.name != "any", "any is a reserved name"

    if args.name == "here":
        assert args.root is None, "root is not allowed"
        args.root = pathlib.Path("../..")
    else:
        assert args.root is not None, "root is required"
        storage = yaml.read(sdir / "storage.yaml")
        assert args.name not in storage, f"storage location '{args.name}' already exists"
        if not args.root.is_absolute() and not args.ssh:
            args.root = pathlib.Path(os.path.relpath(args.root.absolute(), sdir / "storage"))
        if args.mount is not None:
            if not args.mount.is_absolute():
                args.mount = pathlib.Path(os.path.relpath(args.mount.absolute(), sdir / "storage"))

    with search.cwd(sdir):
        loc = Location(root=args.root, ssh=args.ssh, mount=args.mount, prefix=args.prefix)
        s = []
        d = {}
        if args.skip is not None:
            d["skip"] = args.skip
        if args.rglob:
            s.append({"rglob": args.rglob, **d})
        if args.glob:
            s.append({"glob": args.glob, **d})
        if args.exec:
            s.append({"exec": args.exec, **d})
        if len(s) > 0:
            loc.search = s

        loc.overwrite_yaml(f"storage/{args.name}.yaml")

        if args.name != "here":
            yaml.overwrite("storage.yaml", storage + [args.name])

            if args.root.is_absolute() and not args.ssh:
                pathlib.Path(f"data/{args.name}").symlink_to(args.root)
            elif args.ssh is not None and args.mount is not None:
                pathlib.Path(f"data/{args.name}").symlink_to(args.mount)
            elif not args.ssh:
                pathlib.Path(f"data/{args.name}").symlink_to(args.root)
            else:
                pathlib.Path(f"data/{args.name}").symlink_to(pathlib.Path("..") / "unavailable")

        opts = [args.name]
        if args.shallow:
            opts.append("--shallow")
        if args.quiet:
            opts.append("--quiet")
        update(opts)


def _remove_parser():
    """
    Return parser for :py:func:`shelephant rm`.
    """

    desc = textwrap.dedent(
        """
        Remove a storage location to the database.
        The database in ``.shelephant`` is updated as follows:

        -   The ``name`` is removed from ``.shelephant/storage.yaml``.
        -   ``.shelephant/storage/<name>.yaml`` is removed.
        -   The symlink ``.shelephant/data/<name>`` is removed.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("name", type=str, help="Name of the storage location.")
    parser.add_argument("--version", action="version", version=version)
    return parser


def remove(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _remove_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    assert not (sdir / "lock.txt").exists(), "cannot remove location from storage location"

    storage = yaml.read(sdir / "storage.yaml")
    assert args.name in storage, f"storage location '{args.name}' does not exist"
    storage.remove(args.name)
    yaml.overwrite(sdir / "storage.yaml", storage)
    os.remove(sdir / "storage" / f"{args.name}.yaml")
    (sdir / "data" / args.name).unlink()
    update([])


def _update_parser():
    """
    Return parser for :py:func:`shelephant update`.
    """

    desc = textwrap.dedent(
        """
        Update the database.
        This function always update the symbolic links, and optionally updates the available
        files and checksums of (a) )storage location(s).
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("--shallow", action="store_true", help="Do not compute checksums.")
    parser.add_argument("--verbose", action="store_true", help="Verbose commands.")
    parser.add_argument(
        "--chunk",
        type=lambda x: int(float(x)),
        default=1e10,
        help="Chunk size for computing checksums (bytes).",
    )
    parser.add_argument("--force", action="store_true", help="Force update of path(s).")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("name", type=str, nargs="?", help="Update storage location(s).")
    parser.add_argument(
        "path", type=pathlib.Path, nargs="*", help="Update only specific paths on location."
    )
    return parser


def update(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _update_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    base = sdir.parent
    paths = [os.path.relpath(path, base) for path in args.path]
    paths = np.unique(paths) if len(paths) > 0 else None
    lock = None if not (sdir / "lock.txt").exists() else (sdir / "lock.txt").read_text().strip()
    if args.force:
        assert paths is not None, "--force can only be used with path(s)"

    if args.name is None:
        args.name = []
    elif args.name.lower() == "all":
        assert lock is None, "cannot update all locations from storage location"
        args.name = yaml.read(sdir / "storage.yaml")
    else:
        assert lock is None, "cannot update all locations from storage location"
        assert args.name in yaml.read(sdir / "storage.yaml"), f"'{args.name}' is not a location"
        assert args.name != "here" or paths is None, "cannot specify paths for 'here'"
        args.name = [args.name]

    if lock is not None:
        args.name = [lock]

    with search.cwd(sdir):
        symlinks = yaml.read("symlinks.yaml", [])
        symlinks = {pathlib.Path(i["path"]): pathlib.Path(i["storage"]) for i in symlinks}

        # update files and info

        for name in args.name:
            # "here": search for files that are not managed by shelephant
            if name == "here" and paths is None:
                f = "storage/here.yaml"
                Location.from_yaml(f).read().remove(list(symlinks.keys())).to_yaml(f, force=True)
                continue

            # other locations: search for files (or add files by hand)
            loc = Location.from_yaml(f"storage/{name}.yaml")
            if lock is not None:
                loc.root = pathlib.Path("..")
                loc._absroot = loc.root.absolute()
                loc.ssh = None
                loc.mount = False
            if not loc.isavailable():
                continue
            if paths is None:
                loc.read(verbose=args.verbose)
            else:
                _, i, _ = np.intersect1d(paths, loc._files, return_indices=True, assume_unique=True)
                k = np.ones(paths.size, dtype=bool)
                k[i] = False
                loc._append(paths[k])

            if lock is not None:
                f = f"storage/{name}.yaml"
                Location.from_yaml(f).copy_files(loc).overwrite_yaml(f)
            else:
                loc.overwrite_yaml(f"storage/{name}.yaml")

            if args.shallow:
                continue

            # compute sha256/size/mtime of files that changed since the last update
            n = loc._files.size
            if args.force:
                loc.remove_info(paths=paths)
            else:
                loc.check_changes(paths=paths, verbose=args.verbose)

            if loc.has_info():
                if loc._files.size < n:
                    if lock is not None:
                        f = f"storage/{name}.yaml"
                        Location.from_yaml(f).copy_files(loc).overwrite_yaml(f)
                    else:
                        loc.overwrite_yaml(f"storage/{name}.yaml")
                continue

            off = np.sum(loc._size[loc._has_info])
            pbar = tqdm.tqdm(
                total=np.sum(loc._size) - off, disable=args.quiet, unit="B", unit_scale=True
            )
            while not loc.has_info():
                loc.getinfo(
                    paths=paths,
                    max_size=args.chunk,
                    progress=not args.quiet,
                    verbose=args.verbose,
                )
                if lock is not None:
                    f = f"storage/{name}.yaml"
                    Location.from_yaml(f).copy_files(loc).overwrite_yaml(f)
                else:
                    loc.overwrite_yaml(f"storage/{name}.yaml")
                pbar.n = np.sum(loc._size[loc._has_info]) - off
                pbar.refresh()

        # update symlinks
        if lock is not None:
            return

        storage = yaml.read(sdir / "storage.yaml")
        storage.remove("here")
        files = {}
        # - link to first available
        for name in storage[::-1]:
            loc = Location.from_yaml(pathlib.Path("storage") / f"{name}.yaml")
            prefix = loc.prefix if loc.prefix is not None else pathlib.Path(".")
            if loc.isavailable(mount=True):
                for f in loc.files(info=False):
                    if (loc._absroot / f).is_file():
                        files[prefix / pathlib.Path(f)] = pathlib.Path("data") / name
        # - unlinked files: link to first unavailable
        for name in storage[::-1]:
            loc = Location.from_yaml(pathlib.Path("storage") / f"{name}.yaml")
            prefix = loc.prefix if loc.prefix is not None else pathlib.Path(".")
            if not loc.isavailable(mount=True):
                for f in loc.files(info=False):
                    if prefix / pathlib.Path(f) not in files:
                        files[prefix / pathlib.Path(f)] = pathlib.Path("data") / name

        add_links = []
        rm_links = []

        for symlink in symlinks:
            if symlink not in files:
                rm_links.append(symlink)
            elif files[symlink] != symlinks[symlink]:
                add_links.append(symlink)
                rm_links.append(symlink)

        for symlink in files:
            if symlink not in symlinks:
                add_links.append(symlink)

        with search.cwd(sdir / ".."):
            for f in rm_links:
                if f.is_symlink():
                    f.unlink()

            unmanage = []
            for f in add_links:
                if f.is_file():
                    unmanage.append(f)
            for f in unmanage:
                add_links.pop(f)
                files.pop(f)

            if len(unmanage) > 0:
                print("Local files conflicting with dataset. No links are created for these files:")
                print("\n".join(map(str, unmanage)))

            # remove directories that are empty after removing old links
            for d in {i.parent for i in rm_links}:
                if not os.listdir(d):
                    d.rmdir()

            store = [{"path": str(i), "storage": str(files[i])} for i in sorted(files.keys())]
            yaml.overwrite(".shelephant/symlinks.yaml", store)

            for f in add_links:
                f.parent.mkdir(parents=True, exist_ok=True)
                s = pathlib.Path(os.path.relpath(".shelephant", f.parent)) / files[f] / f
                f.symlink_to(s)


def _cp_parser():
    """
    Return parser for :py:func:`shelephant cp`.
    """

    desc = textwrap.dedent(
        """
        Copy files between storage locations.

        .. note::

            The copied files are added to the database of the destination.
            There is no check that this fits ``dump`` and ``search`` settings.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("--colors", type=str, default="dark", help="Color scheme [none, dark].")
    parser.add_argument("-f", "--force", action="store_true", help="Overwrite without prompt.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("-n", "--dry-run", action="store_true", help="Print copy-plan and exit.")
    parser.add_argument("source", type=str, help="name of the source.")
    parser.add_argument("destination", type=str, help="name of the destination.")
    parser.add_argument("path", type=pathlib.Path, nargs="+", help="path(s) to copy.")
    return parser


def cp(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _cp_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    assert not (sdir / "lock.txt").exists(), "cannot remove location from storage location"
    assert args.destination != "here", "Cannot copy to here."
    storage = yaml.read(sdir / "storage.yaml")
    assert args.source in storage, f"Unknown storage location {args.source}"
    assert args.destination in storage, f"Unknown storage location {args.destination}"
    base = sdir.parent
    paths = [os.path.relpath(path, base) for path in args.path]

    with search.cwd(sdir):
        opts = [f"storage/{args.source}.yaml", f"storage/{args.destination}.yaml"]
        opts += ["--colors", args.colors]
        opts += ["--force"] if args.force else []
        opts += ["--quiet"] if args.quiet else []
        opts += ["--dry-run"] if args.dry_run else []
        changed = cli.shelephant_cp(opts, paths)

    if not args.dry_run and len(changed) > 0:
        _, j, _ = np.intersect1d(paths, changed, return_indices=True, assume_unique=True)
        changed = np.array(args.path)[j]
        update(["--quiet", "--force", args.destination] + list(map(str, changed)))


def _mv_parser():
    """
    Return parser for :py:func:`shelephant mv`.
    """

    desc = textwrap.dedent(
        """
        Move files from one storage location to another.

        .. note::

            The copied files are added to the database of the destination.
            There is no check that this fits ``dump`` and ``search`` settings.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("--colors", type=str, default="dark", help="Color scheme [none, dark].")
    parser.add_argument("-f", "--force", action="store_true", help="Overwrite without prompt.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("-n", "--dry-run", action="store_true", help="Print copy-plan and exit.")
    parser.add_argument("source", type=str, help="name of the source.")
    parser.add_argument("destination", type=str, help="name of the destination.")
    parser.add_argument("path", type=pathlib.Path, nargs="+", help="path(s) to copy.")
    return parser


def mv(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _mv_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    assert not (sdir / "lock.txt").exists(), "cannot remove location from storage location"
    assert args.destination != "here", "Cannot copy to here."
    storage = yaml.read(sdir / "storage.yaml")
    assert args.source in storage, f"Unknown storage location {args.source}"
    assert args.destination in storage, f"Unknown storage location {args.destination}"
    base = sdir.parent
    paths = [os.path.relpath(path, base) for path in args.path]

    with search.cwd(sdir):
        dest = Location.from_yaml(f"storage/{args.destination}.yaml")
        assert dest.ssh is None, "Cannot move to remote location."
        opts = [f"storage/{args.source}.yaml", str(dest._absroot)]
        opts += ["--colors", args.colors]
        opts += ["--force"] if args.force else []
        opts += ["--quiet"] if args.quiet else []
        opts += ["--dry-run"] if args.dry_run else []
        cli.shelephant_mv(opts, paths)

    if not args.dry_run:
        with search.cwd(sdir):
            f = f"storage/{args.source}.yaml"
            Location.from_yaml(f).remove(paths).overwrite_yaml(f)
        update(["--quiet", "--force", args.destination] + list(map(str, args.path)))


def _rm_parser():
    """
    Return parser for :py:func:`shelephant rm`.
    """

    desc = textwrap.dedent(
        """
        Remove files from a storage location.

        .. warning::

            This remove the actual data (and the link, if there is no alternative source left).
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("-f", "--force", action="store_true", help="Overwrite without prompt.")
    parser.add_argument("-q", "--quiet", action="store_true", help="Do not print progress.")
    parser.add_argument("-n", "--dry-run", action="store_true", help="Print copy-plan and exit.")
    parser.add_argument("source", type=str, help="name of the source.")
    parser.add_argument("path", type=pathlib.Path, nargs="+", help="path(s) to remove.")
    return parser


def rm(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _rm_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    assert not (sdir / "lock.txt").exists(), "cannot remove location from storage location"
    storage = yaml.read(sdir / "storage.yaml")
    assert args.source in storage, f"Unknown storage location {args.source}"
    base = sdir.parent
    paths = [os.path.relpath(path, base) for path in args.path]

    with search.cwd(sdir):
        opts = [f"storage/{args.source}.yaml"]
        opts += ["--force"] if args.force else []
        opts += ["--quiet"] if args.quiet else []
        opts += ["--dry-run"] if args.dry_run else []
        cli.shelephant_rm(opts, paths)

    if not args.dry_run:
        with search.cwd(sdir):
            f = f"storage/{args.source}.yaml"
            Location.from_yaml(f).remove(paths).overwrite_yaml(f)
            update([])


def _pwd_parser():
    """
    Return parser for :py:func:`shelephant pwd`.
    """

    desc = textwrap.dedent(
        """
        Change the current working directory to a storage location.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("--base", action="store_true", help="Print the base directory.")
    parser.add_argument("--abspath", action="store_true", help="Print absolute path.")
    parser.add_argument("source", type=str, help="name of the source.")
    return parser


def pwd(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _pwd_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    assert not (sdir / "lock.txt").exists(), "not available from storage location"
    storage = yaml.read(sdir / "storage.yaml")
    assert args.source in storage, f"Unknown storage location {args.source}"

    cwd = pathlib.Path.cwd()
    post = os.path.relpath(cwd, sdir / "..")

    with search.cwd(sdir):
        f = f"storage/{args.source}.yaml"
        loc = Location.from_yaml(f)
        prefix = loc.prefix
        root = sdir / "storage" / loc.root

    if prefix is not None:
        common = os.path.commonprefix([prefix, post])
        post = post[len(common) :]

    if args.base:
        post = ""

    if args.abspath:
        print(os.path.normpath(root / post))
    else:
        print(os.path.relpath(root / post, os.getcwd()))


def _status_parser():
    """
    Return parser for :py:func:`shelephant status`.
    """

    desc = textwrap.dedent(
        """
        Status of the storage locations.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("--min-copies", type=int, help="Show files with minimal number of copies.")
    parser.add_argument("--copies", type=int, help="Show files with specific number of copies.")
    parser.add_argument("--ne", action="store_true", help="Show files with unequal copies.")
    parser.add_argument("--na", action="store_true", help="Show files unavailable somewhere.")
    parser.add_argument("--unknown", action="store_true", help="Show files with unknown sha256.")
    parser.add_argument("--table", type=str, default="SINGLE_BORDER", help="Select print style.")
    parser.add_argument("--in-use", type=str, help="Select storage location in use.")
    parser.add_argument("--output", type=pathlib.Path, help="Dump to YAML file.")
    parser.add_argument("--copy", type=str, nargs=2, help="Copy file selection.")
    parser.add_argument(
        "-b",
        "--relative-to-base",
        action="store_true",
        help="Show path relative to base directory of dataset.",
    )
    parser.add_argument(
        "path",
        type=str,
        nargs="*",
        help="Filter to paths (either one directory, or multiple files).",
    )
    return parser


def status(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _status_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    base = sdir.parent
    cwd = os.path.relpath(pathlib.Path.cwd(), base)
    paths = [os.path.relpath(path, base) for path in args.path]

    if args.output is not None:
        raise NotImplementedError

    if args.copy is not None:
        raise NotImplementedError

    with search.cwd(sdir):
        symlinks = np.sort([i["path"] for i in yaml.read("symlinks.yaml", [])])
        storage = yaml.read(sdir / "storage.yaml")
        storage.remove("here")
        extra = Location.from_yaml("storage/here.yaml").files(info=False)

        sha = "x" * np.ones((len(symlinks), len(storage)), dtype=object)
        mtime = np.inf * np.ones(sha.shape, dtype=np.float64)
        inuse = "----" * np.ones((len(symlinks)), dtype=object)

        for iname, name in enumerate(storage[::-1]):
            loc = Location.from_yaml(pathlib.Path("storage") / f"{name}.yaml")
            if loc.prefix is not None:
                files = np.array([str(loc.prefix / f) for f in loc._files])
            else:
                files = loc._files
            sorter = np.argsort(files)
            idx = np.searchsorted(symlinks, files[sorter])
            s = np.array(loc._sha256)[sorter]
            m = np.array(loc._mtime)[sorter]
            h = ~loc._has_info[sorter]
            if np.any(h):
                s[h] = "?"
                m[h] = np.inf
            sha[idx, -1 - iname] = s
            mtime[idx, -1 - iname] = m
            if loc.isavailable(mount=True):
                inuse[idx] = name

    for i in range(sha.shape[0]):
        _sha = sha[i, :]
        unique = np.unique(_sha)
        info = np.logical_and(unique != "x", unique != "?")

        if np.sum(info) == 0:
            continue

        if np.sum(info) == 1:
            _sha[_sha == unique[info][0]] = "=="
            sha[i, :] = _sha
            continue

        info = np.logical_and(_sha != "x", _sha != "?")
        _mtime = mtime[i, :]
        _mtime[~info] = np.inf
        sorter = np.array(list(map(str, 1 + np.argsort(_mtime))), dtype=object)
        _sha[info] = sorter[info]
        sha[i, :] = _sha

    data = np.hstack((np.array([symlinks]).T, inuse.reshape(-1, 1), sha))

    e = "x" * np.ones((len(extra), data.shape[1]), dtype=object)
    e[:, 0] = extra
    e[:, 1] = "here"
    data = np.vstack((data, e))

    if args.min_copies is not None:
        data = data[np.sum(data[:, 2:] == "==", axis=1) >= args.min_copies]
    if args.copies is not None:
        data = data[np.sum(data[:, 2:] == "==", axis=1) == args.copies]
    if args.ne:
        data = data[np.sum(data[:, 2:] == "1", axis=1) > 0]
    if args.na:
        data = data[np.sum(data[:, 2:] == "x", axis=1) > 0]
    if args.unknown:
        data = data[np.sum(data[:, 2:] == "?=", axis=1) > 0]
    if args.in_use is not None:
        data = data[data[:, 1] == args.in_use]

    if len(paths) > 0:
        keep = np.zeros(data.shape[0], dtype=bool)
        for path in paths:
            keep = np.logical_or(
                keep, [os.path.commonpath([str(i), path]) == path for i in data[:, 0]]
            )
        data = data[keep]

    if not args.relative_to_base:
        data[:, 0] = [os.path.relpath(i, cwd) for i in data[:, 0]]

    out = prettytable.PrettyTable()
    if args.table == "PLAIN_COLUMNS":
        out.set_style(prettytable.PLAIN_COLUMNS)
    elif args.table == "SINGLE_BORDER":
        out.set_style(prettytable.SINGLE_BORDER)

    out.field_names = ["path", "in use"] + storage

    out.align["path"] = "l"
    out.align["in use"] = "l"
    for name in storage:
        out.align[name] = "c"

    for row in data:
        out.add_row(row)

    print(out.get_string())


def _info_parser():
    """
    Return parser for :py:func:`shelephant info`.
    """

    desc = textwrap.dedent(
        """
        Show global information about dataset.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    parser.add_argument("--cachedir", action="store_true", help="Print cachedir.")
    parser.add_argument(
        "--basedir", action="store_true", help="Print basedir (that contain '.shelephant')."
    )
    return parser


def info(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _info_parser()
    args = parser.parse_args(args)

    if args.cachedir:
        print(user_cache_dir("shelephant", "tdegeus"))
        return

    if args.basedir:
        print(_search_upwards_dir(".shelephant").parent)
        return


def _find_matching(
    text: str,
    opening: str,
    closing: str,
) -> dict:
    r"""
    Find matching 'brackets'.

    :param text: The string to consider.
    :param opening: The opening bracket (e.g. "(", "[", "{").
    :param closing: The closing bracket (e.g. ")", "]", "}").
    :return: Dictionary with ``{index_opening: index_closing}``
    """

    opening = [i.span()[0] for i in re.finditer(opening, text)]
    closing = [i.span()[1] for i in re.finditer(closing, text)]

    if len(opening) == 0:
        return {}

    if len(opening) > len(closing):
        raise IndexError("Unmatching opening...closing found")

    opening = np.array(opening, dtype=int)
    closing = np.array(closing, dtype=int) * -1
    brackets = np.concatenate([opening, closing])
    brackets = brackets[np.argsort(np.abs(brackets))]

    ret = {}
    stack = []

    for i in brackets:
        if i >= 0:
            stack.append(i)
        else:
            if len(stack) == 0:
                raise IndexError(f"No closing {closing} at: {i:d}")
            j = stack.pop()
            ret[j] = -1 * i

    if len(stack) > 0:
        i = stack.pop()
        raise IndexError(f"No opening {opening} at: {i:d}")

    return ret


def _linenumbers(text: str) -> list[int]:
    """
    Return the line-number of each character in ``text``.

    :param text: The text to consider.
    :return: List of line-numbers.
    """
    lineno = np.empty(len(text), dtype=int)
    i = 0
    line = 0
    for line, match in enumerate(re.finditer(r"\n", text)):
        lineno[i : match.span()[0]] = line
        i = match.span()[0]
        lineno[i] = line
        i += 1
    lineno[i:] = line + 1
    lineno = np.append(lineno, line + 2)
    return lineno


def _gitignore_parser():
    """
    Return parser for :py:func:`shelephant gitignore`.
    """

    desc = textwrap.dedent(
        """
        Add all symbolic links at ``.shelephant`` to dataset's ``.gitignore``.
        Note that this is not ``.shelephant/.gitignore`` but the one in the dataset's root.
        """
    )

    class MyFmt(
        argparse.RawDescriptionHelpFormatter,
        argparse.ArgumentDefaultsHelpFormatter,
        argparse.MetavarTypeHelpFormatter,
    ):
        pass

    parser = argparse.ArgumentParser(formatter_class=MyFmt, description=desc)

    parser.add_argument("--version", action="version", version=version)
    return parser


def gitignore(args: list[str]):
    """
    Command-line tool, see ``--help``.

    :param args: Command-line arguments (should be all strings).
    """

    parser = _gitignore_parser()
    args = parser.parse_args(args)
    sdir = _search_upwards_dir(".shelephant")
    assert sdir is not None, "Not a shelephant dataset"
    path = sdir / ".." / ".gitignore"

    if path.exists():
        ignore = path.read_text()
        brackets = _find_matching(ignore, re.escape("# <shelephant>"), re.escape("# </shelephant>"))
        if len(brackets) > 0:
            assert len(brackets) == 1, "Multiple shelephant sections found"
            brackets = [(key, value) for key, value in brackets.items()][0]
            lineno = _linenumbers(ignore)
            ignore = ignore.splitlines()
            ignore = (
                "\n".join(ignore[: lineno[brackets[0]]]).rstrip()
                + "\n"
                + "\n".join(ignore[lineno[brackets[1]] + 1 :]).lstrip()
            )
    else:
        ignore = ""

    with search.cwd(sdir):
        symlinks = yaml.read("symlinks.yaml", [])
        symlinks = [i["path"] for i in symlinks]

    if len(symlinks) > 0:
        ignore += "\n# <shelephant>\n" + "\n".join(symlinks) + "\n# </shelephant>\n"
    path.write_text(ignore.strip())


def git(args: list[str]):
    """
    Run git from ``.shelephant`` directory.

    :param args: Command-line arguments (should be all strings).
    """
    with search.cwd(_search_upwards_dir(".shelephant")):
        print(exec_cmd(f"git {' '.join(args)}"))



import os
import pathlib

import click
import yaml

from . import convert


def read(filename: str | pathlib.Path, default=None) -> list | dict:
    r"""
    Read YAML file and return its content.

    :param filename: The YAML file to read.
    :param default: The default value to return if the file is empty.
    :return: The content of the YAML file.
    """

    if not os.path.isfile(filename):
        raise OSError(f'"{filename} does not exist')

    with open(filename) as file:
        ret = yaml.load(file.read(), Loader=yaml.FullLoader)
        if ret is None:
            return default
        return ret


def read_item(filename: str | pathlib.Path, key: str | list[str] = []) -> list | dict:
    r"""
    Get an item from a YAML file.

    :param key:
        The item to read. E.g.
        *   ``[]`` for a YAML file containing only a list.
        *   ``['foo']`` for a plain YAML file.
        *   ``['key', 'to', foo']`` for a YAML file with nested items.

        An item specified as ``str`` separated by "/" is also accepted.

    :return: The content of the item.
    """

    data = read(filename)

    if type(data) == dict:
        return convert.get(data, key)

    key = convert.split_key(key)

    if type(data) == list and len(key) == 0:
        return data

    raise OSError(f'"{"/".join(key)}" not in "{filename}"')


def dump(filename: str | pathlib.Path, data: list | dict, force: bool = False):
    """
    Dump data to YAML file.

    :param filename: The output filename.
    :param data: The data to dump.
    :param force: Do not prompt to overwrite file.
    """

    dirname = os.path.dirname(filename)

    if not force:
        if os.path.isfile(filename):
            if not click.confirm(f'Overwrite "{filename}"?'):
                raise OSError("Cancelled")
        elif not os.path.isdir(dirname) and len(dirname) > 0:
            if not click.confirm(f'Create "{os.path.dirname(filename)}"?'):
                raise OSError("Cancelled")

    if not os.path.isdir(dirname) and len(dirname) > 0:
        os.makedirs(os.path.dirname(filename))

    with open(filename, "w") as file:
        yaml.dump(data, file)


def overwrite(filename: str | pathlib.Path, data: list | dict):
    """
    Overwrite existing YAML file with data.
    This function only changes the file if the content has indeed changed.

    :param filename: The output filename.
    :param data: The data to dump.
    """

    if not os.path.isfile(filename):
        return dump(filename, data)

    ret = yaml.dump(data, default_flow_style=False, default_style="")
    old = pathlib.Path(filename).read_text()

    if ret == old:
        return

    pathlib.Path(filename).write_text(ret)


def preview(data: list | dict):
    r"""
    Print data formatted as YAML.

    :param data: The data to dump.
    """
    print(yaml.dump(data, default_flow_style=False, default_style=""))



import os

import tqdm

from .external import exec_cmd


def copy(
    source_dir: str,
    dest_dir: str,
    files: list[str],
    options: str = "-p",
    verbose: bool = False,
    progress: bool = True,
):
    """
    Copy files using *scp*.

    :param source_dir: Source directory. If remote: ``[user@]host:path``.
    :param dest_dir: Source directory. If remote: ``[user@]host:path``.
    :param files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param options: Options passed to ``scp``.
    :param verbose: Verbose commands.
    :param progress: Show progress bar.
    """

    for file in tqdm.tqdm(files, disable=not progress):
        cmd = f"scp {options:s} {os.path.join(source_dir, file):s} {os.path.join(dest_dir, file):s}"
        exec_cmd(cmd, verbose)



import os
from collections import defaultdict

import click


def _to_tree(d):
    r"""
    Detail for: :py:fun:`filter_deepest`.
    Not part of API.
    """

    t = defaultdict(list)
    for a, *b in d:
        t[a].append(b)
    return {a: None if not (k := list(filter(None, b))) else _to_tree(k) for a, b in t.items()}


def _get_deepest_paths(d, c=[]):
    r"""
    Detail for: :py:fun:`filter_deepest`.
    Not part of API.

    See https://stackoverflow.com/a/66211932/2646505
    """
    for a, b in d.items():
        if b is None:
            yield "/".join(c + [a])
        else:
            yield from _get_deepest_paths(b, c + [a])


def filter_deepest(files: list[str]) -> list[str]:
    r"""
    Return list with only the deepest paths.

    For example::

        filter_deepest(["foo/bar/dir", "foo/bar"])
        >>> ["foo/bar/dir"]

    :param files: List of paths.
    :return: List of paths.
    """

    return list(_get_deepest_paths(_to_tree([i.split("/") for i in files])))


def dirnames(files: list[str], return_unique: bool = True) -> list[str]:
    r"""
    Get the ``os.path.dirname`` of all file paths.

    :param files: List of file paths.
    :param return_unique: Filter duplicates.
    :return: List of dirnames.
    """

    if type(files) == str:
        files = [files]

    ret = [os.path.dirname(filename) for filename in files]

    if not return_unique:
        return ret

    return list(set(ret))


def makedirs(dirnames: list[str], force: bool = False):
    r"""
    (Prompt and) Create directories that do not yet exist.
    This function creates parent directories if needed.

    :param dirnames: List of directory paths.
    :param force: Create directories without prompt.
    """

    if type(dirnames) == str:
        dirnames = [dirnames]

    dirnames = [dirname for dirname in dirnames if len(dirname) > 0]
    dirnames = [dirname for dirname in dirnames if not os.path.isdir(dirname)]

    if len(dirnames) == 0:
        return 0

    dirnames = list(set(dirnames))
    dirnames = sorted(filter_deepest(dirnames))

    if not force:
        for dirname in dirnames:
            print(f"mkdir -p {dirname:s}")
        if not click.confirm("Proceed?"):
            raise OSError("Cancelled")

    for dirname in dirnames:
        os.makedirs(dirname)



import subprocess


def exec_cmd(cmd, verbose=False):
    r"""
    Run command, optionally verbose command and its output, and return output.

    :type cmd: str
    :param cmd: The command to run.

    :type verbose: bool
    :param verbose: Print command and its output.
    """

    if verbose:
        print(cmd)

    ret = subprocess.check_output(cmd, shell=True).decode("utf-8")

    if verbose:
        print(ret)

    return ret



import re
import unittest

import shelephant


class Test_path(unittest.TestCase):
    def test_filter_deepest(self):
        check = ["/foo/bar/dir/a", "/foo/bar/dir/b"]
        files = check + ["/foo/bar", "/foo/bar/dir"]
        ret = shelephant.path.filter_deepest(files)
        self.assertEqual(ret, check)

    def test_filter_deepest_2(self):
        dirnames = [
            "foo/bar",
            "foo/bar/mydir",
            "bar/mydir",
            "bar",
            "bar/mydir2",
            "foo/bar/mydir/deep",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow",
        ]

        ret = shelephant.path.filter_deepest(dirnames)

        d = [
            "bar/mydir",
            "bar/mydir2",
            "foo/bar/mydir/deep/deeper",
            "foo/bar/mydir/deep/also",
            "foo/shallow",
        ]

        self.assertEqual(sorted(ret), sorted(d))

    def test_dirname(self):
        files = ["/foo/bar", "/foo/dir/bar"]
        ret = shelephant.path.dirnames(files)
        self.assertEqual(sorted(ret), ["/foo", "/foo/dir"])


class Test_convert(unittest.TestCase):
    def test_flatten(self):
        arg = [1, [2, 2, 2], 4]
        ret = [1, 2, 2, 2, 4]

        self.assertEqual(ret, shelephant.convert.flatten(arg))

    def test_squash(self):
        arg = {"foo": [1, 2], "bar": {"foo": [3, 4], "bar": 5}}
        ret = [1, 2, 3, 4, 5]

        self.assertEqual(ret, shelephant.convert.squash(arg))


class Test_output(unittest.TestCase):
    def test_copyplan(self):
        ret = shelephant.output.copyplan(
            status={"==": ["skip.txt"], "->": ["right.txt"], "!=": ["overwrite.txt"]},
            display=False,
        )
        expect = [
            "overwrite.txt => overwrite.txt",
            "right.txt -> right.txt",
            "skip.txt == skip.txt",
        ]
        ret = [re.sub(r"\s+", " ", line) for line in ret.splitlines()]
        self.assertEqual(ret, expect)


if __name__ == "__main__":
    unittest.main()



import os
import pathlib
import shutil

import numpy as np
import tqdm


def remove(
    source_dir: str,
    files: list[str],
    progress: bool = True,
):
    """
    Remove files using ``os.remove``.

    :param source_dir: Source directory
    :param dest_dir: Source directory
    :param files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param progress: Show progress bar.
    """

    for file in tqdm.tqdm(files, disable=not progress):
        os.remove(os.path.join(source_dir, file))


def move(
    source_dir: str,
    dest_dir: str,
    files: list[str],
    progress: bool = True,
):
    """
    Move files using ``os.replace``.

    :param source_dir: Source directory
    :param dest_dir: Source directory
    :param files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param progress: Show progress bar.
    """

    for file in tqdm.tqdm(files, disable=not progress):
        s = os.path.join(source_dir, file)
        d = os.path.join(dest_dir, file)
        pathlib.Path(d).parent.mkdir(parents=True, exist_ok=True)
        os.replace(s, d)


def copy(
    source_dir: str,
    dest_dir: str,
    files: list[str],
    progress: bool = True,
):
    """
    Copy files using ``shutil.copy2``.

    :param source_dir: Source directory
    :param dest_dir: Source directory
    :param files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param progress: Show progress bar.
    """

    for file in tqdm.tqdm(files, disable=not progress):
        s = os.path.join(source_dir, file)
        d = os.path.join(dest_dir, file)
        pathlib.Path(d).parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(s, d)


def diff(
    source_dir: str,
    dest_dir: str,
    files: list[str],
) -> dict[list[str]]:
    """
    Check if files exist.

    :param str source_dir: Source directory (optionally with hostname).
    :param str dest_dir: Source directory (optionally with hostname).
    :param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param verbose: Verbose commands.
    :return:
        Dictionary with differences::

            {
                "?=" : [ ... ], # in source_dir and dest_dir
                "->" : [ ... ], # in source_dir not in dest_dir
                "<-" : [ ... ], # in dest_dir not in source_dir
            }
    """

    insource = np.array([os.path.exists(os.path.join(source_dir, file)) for file in files])
    indest = np.array([os.path.exists(os.path.join(dest_dir, file)) for file in files])
    files = np.array(files)

    return {
        "?=": files[np.logical_and(insource, indest)].tolist(),
        "->": files[np.logical_and(insource, ~indest)].tolist(),
        "<-": files[np.logical_and(~insource, indest)].tolist(),
    }



import contextlib
import io
import os
import pathlib
import re
import shutil
import unittest

import shelephant
from shelephant._tests import create_dummy_files
from shelephant.cli import f_dump
from shelephant.cli import f_hostinfo
from shelephant.cli import shelephant_cp
from shelephant.cli import shelephant_diff
from shelephant.cli import shelephant_dump
from shelephant.cli import shelephant_hostinfo
from shelephant.cli import shelephant_mv
from shelephant.cli import shelephant_parse
from shelephant.cli import shelephant_rm
from shelephant.search import cwd
from shelephant.search import tempdir

has_ssh = shelephant.ssh.has_keys_set("localhost")
has_rsync = shutil.which("rsync") is not None


def _plain(text):
    return list(filter(None, [re.sub(r"\s+", " ", line).strip() for line in text.splitlines()]))


class Test_shelephant_parse(unittest.TestCase):
    def test_basic(self):
        """
        shelephant_parse <file.yaml>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            files = ["foo.txt", "bar.txt"]
            create_dummy_files(files)
            shelephant_dump(["-i"] + files)
            check = pathlib.Path(f_dump).read_text()
            shelephant_parse([f_dump])

        self.assertEqual(sio.getvalue().strip(), check.strip())


class Test_shelephant_dump(unittest.TestCase):
    def test_checksum(self):
        with tempdir():
            files = ["foo.txt", "bar.txt"]
            check = create_dummy_files(files)
            shelephant_dump(["-i"] + files)
            data = shelephant.dataset.Location.from_yaml(f_dump)
            self.assertTrue(check == data)

    def test_search(self):
        with tempdir():
            files = ["foo.txt", "bar.txt", "a.txt", "b.txt", "c.txt", "d.txt"]
            check = create_dummy_files(files)
            shelephant.yaml.dump("search.yaml", {"root": ".", "search": [{"rglob": "*.txt"}]})
            shelephant_dump(["-i", "--search", "search.yaml"])
            data = shelephant.dataset.Location.from_yaml(f_dump)
            self.assertTrue(check == data)

    def test_find(self):
        with tempdir():
            files = ["foo.txt", "bar.txt", "a.txt", "b.txt", "c.txt", "d.txt"]
            check = create_dummy_files(files)
            shelephant_dump(["-i", "-c", "find . -name '*.txt'"])
            data = shelephant.dataset.Location.from_yaml(f_dump)
            self.assertTrue(check == data)

    def test_append(self):
        with tempdir():
            files = ["foo.pdf", "bar.pdf", "a.txt", "b.txt", "c.txt", "d.txt"]
            check = create_dummy_files(files)
            pdf = [i for i in files if i.endswith(".pdf")]

            # plain

            shelephant_dump(pdf)
            data = shelephant.dataset.Location.from_yaml(f_dump)
            self.assertEqual(data.files(info=False), pdf)

            shelephant_dump(["-a", "-c", "find . -name '*.txt'"])
            data = shelephant.dataset.Location.from_yaml(f_dump).sort()
            self.assertEqual(data.files(info=False), sorted(files))

            # with details

            shelephant_dump(["-f", "-i"] + pdf)
            data = shelephant.dataset.Location.from_yaml(f_dump)
            self.assertEqual(data.files(info=False), pdf)

            shelephant_dump(["-a", "-i", "-c", "find . -name '*.txt'"])
            data = shelephant.dataset.Location.from_yaml(f_dump)
            self.assertTrue(check == data)


class Test_shelephant_hostinfo(unittest.TestCase):
    def test_search(self):
        with tempdir():
            files = ["foo.txt", "bar.txt"]
            check = create_dummy_files(files)
            shelephant.yaml.dump("info.yaml", {"root": ".", "search": [{"rglob": "*.txt"}]})
            shelephant_hostinfo(["-iu", "info.yaml"])
            loc = shelephant.dataset.Location.from_yaml("info.yaml")
            self.assertTrue(check == loc)


class Test_shelephant_cp(unittest.TestCase):
    def test_basic(self):
        """
        shelephant_cp <sourceinfo.yaml> <dest_dirname>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(files)
                args = [f_dump, "../dest"]
                shelephant_cp(["-n", "--colors", "none"] + args)
                shelephant_cp(["-f", "--quiet"] + args)

            data = shelephant.dataset.Location(root="dest", files=files).getinfo()
            self.assertTrue(check == data)

        if has_rsync:
            expect = [
                "bar.txt => bar.txt",
                "more.txt -> more.txt",
                "even_more.txt -> even_more.txt",
                "foo.txt == foo.txt",
            ]
        else:
            expect = [
                "foo.txt => foo.txt",
                "bar.txt => bar.txt",
                "more.txt -> more.txt",
                "even_more.txt -> even_more.txt",
            ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_basic_basic(self):
        """
        shelephant_cp <sourceinfo.yaml> <dest_dirname> --mode basic
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(files)
                args = [f_dump, "../dest", "--mode", "basic"]
                shelephant_cp(["-n", "--colors", "none"] + args)
                shelephant_cp(["-f", "--quiet"] + args)

            data = shelephant.dataset.Location(root="dest", files=files).getinfo()
            self.assertTrue(check == data)

        expect = [
            "foo.txt => foo.txt",
            "bar.txt => bar.txt",
            "more.txt -> more.txt",
            "even_more.txt -> even_more.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_destinfo(self):
        """
        shelephant_cp <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(files)
                shelephant_hostinfo(["../dest"])
                args = [f_dump, f_hostinfo]
                shelephant_cp(["-n", "--colors", "none"] + args)
                shelephant_cp(["-f", "--quiet"] + args)

            data = shelephant.dataset.Location(root="dest", files=files).getinfo()
            self.assertTrue(check == data)

        if has_rsync:
            expect = [
                "bar.txt => bar.txt",
                "more.txt -> more.txt",
                "even_more.txt -> even_more.txt",
                "foo.txt == foo.txt",
            ]
        else:
            expect = [
                "foo.txt => foo.txt",
                "bar.txt => bar.txt",
                "more.txt -> more.txt",
                "even_more.txt -> even_more.txt",
            ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_destinfo_sha256(self):
        """
        shelephant_cp <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))
                shelephant_dump(["-i", "foo.txt", "bar.txt"])

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(["-i"] + files)
                shelephant_hostinfo(["../dest", "-d"])
                args = [f_dump, f_hostinfo]
                shelephant_cp(["-n", "--colors", "none"] + args)
                shelephant_cp(["-f", "--quiet"] + args)

            data = shelephant.dataset.Location(root="dest", files=files).getinfo()
            self.assertTrue(check == data)

        expect = [
            "bar.txt => bar.txt",
            "more.txt -> more.txt",
            "even_more.txt -> even_more.txt",
            "foo.txt == foo.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_destinfo_sha256_2(self):
        """
        shelephant_cp <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))
                shelephant_dump(["foo.txt", "bar.txt"])

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(["-i"] + files)
                shelephant_hostinfo(["../dest", "-d", "--info"])
                args = [f_dump, f_hostinfo]
                shelephant_cp(["-n", "--colors", "none"] + args)
                shelephant_cp(["-f", "--quiet"] + args)

            data = shelephant.dataset.Location(root="dest", files=files).getinfo()
            self.assertTrue(check == data)

        expect = [
            "bar.txt => bar.txt",
            "more.txt -> more.txt",
            "even_more.txt -> even_more.txt",
            "foo.txt == foo.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_sourceinfo(self):
        """
        shelephant_cp <sourceinfo.yaml> <dest_dirname>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(files)

            with cwd("dest"):
                shelephant_hostinfo(["../src", "-d"])
                args = [f_hostinfo, "."]
                shelephant_cp(["-n", "--colors", "none"] + args)
                shelephant_cp(["-f", "--quiet"] + args)
                shelephant_dump(["-i", "-c", "find . -name '*.txt'"])
                data = shelephant.dataset.Location.from_yaml(f_dump)

            self.assertTrue(check == data)

        if has_rsync:
            expect = [
                "bar.txt => bar.txt",
                "more.txt -> more.txt",
                "even_more.txt -> even_more.txt",
                "foo.txt == foo.txt",
            ]
        else:
            expect = [
                "foo.txt => foo.txt",
                "bar.txt => bar.txt",
                "more.txt -> more.txt",
                "even_more.txt -> even_more.txt",
            ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_ssh_send(self):
        """
        shelephant_cp <sourceinfo.yaml> <dest_dirname_on_host> --ssh <user@host>
        """
        if not has_ssh:
            raise unittest.SkipTest("'ssh localhost' does not work")

        with tempdir(), shelephant.ssh.tempdir("localhost") as remote, contextlib.redirect_stdout(
            io.StringIO()
        ) as sio:
            files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
            check = create_dummy_files(files)
            shelephant_dump(files)

            args = [f_dump, str(remote), "--ssh", "localhost"]
            shelephant_cp(["-n", "--colors", "none"] + args)
            shelephant_cp(["-f", "--quiet"] + args)
            data = shelephant.dataset.Location(root=remote, files=files).getinfo()

        self.assertTrue(check == data)

        expect = [
            "foo.txt -> foo.txt",
            "bar.txt -> bar.txt",
            "more.txt -> more.txt",
            "even_more.txt -> even_more.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_ssh_send2(self):
        """
        shelephant_cp <sourceinfo.yaml> <remote_destinfo.yaml>
        """
        if not has_ssh:
            raise unittest.SkipTest("'ssh localhost' does not work")

        with tempdir(), shelephant.ssh.tempdir("localhost") as remote, contextlib.redirect_stdout(
            io.StringIO()
        ) as sio:
            files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
            check = create_dummy_files(files)
            shelephant_dump(files)
            shelephant_hostinfo([str(remote), "--ssh", "localhost"])

            args = [f_dump, f_hostinfo]
            shelephant_cp(["-n", "--colors", "none"] + args)
            shelephant_cp(["-f", "--quiet"] + args)
            data = shelephant.dataset.Location(root=remote, files=files).getinfo()

        self.assertTrue(check == data)

        expect = [
            "foo.txt -> foo.txt",
            "bar.txt -> bar.txt",
            "more.txt -> more.txt",
            "even_more.txt -> even_more.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_ssh_get(self):
        """
        shelephant_cp <remote_sourceinfo.yaml> <dest_dirname>
        """
        if not has_ssh:
            raise unittest.SkipTest("'ssh localhost' does not work")

        with tempdir(), shelephant.ssh.tempdir("localhost") as remote, contextlib.redirect_stdout(
            io.StringIO()
        ) as sio:
            with cwd(remote):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(files)

            shelephant_hostinfo([str(remote), "--ssh", "localhost", "-d"])
            args = [f_hostinfo, "."]
            shelephant_cp(["-n", "--colors", "none"] + args)
            shelephant_cp(["-f", "--quiet"] + args)
            data = shelephant.dataset.Location(root=".", files=files).getinfo()

        self.assertTrue(check == data)

        expect = [
            "foo.txt -> foo.txt",
            "bar.txt -> bar.txt",
            "more.txt -> more.txt",
            "even_more.txt -> even_more.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)


class Test_shelephant_mv(unittest.TestCase):
    def test_basic(self):
        """
        shelephant_mv <sourceinfo.yaml> <dest_dirname>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                check = create_dummy_files(files)
                shelephant_dump(files)
                args = [f_dump, "../dest"]
                shelephant_mv(["-n", "--colors", "none"] + args)
                shelephant_mv(["-f", "--quiet"] + args)
                self.assertFalse(any([os.path.exists(f) for f in files]))

            data = shelephant.dataset.Location(root="dest", files=files).getinfo()
            self.assertTrue(check == data)

        expect = [
            "foo.txt => foo.txt",
            "bar.txt => bar.txt",
            "more.txt -> more.txt",
            "even_more.txt -> even_more.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)


class Test_shelephant_rm(unittest.TestCase):
    def test_basic(self):
        """
        shelephant_rm <sourceinfo.yaml>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()

            files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
            create_dummy_files(files)
            shelephant_dump(files)
            args = [f_dump]
            shelephant_rm(["-n"] + args)
            shelephant_rm(["-f", "--quiet"] + args)
            self.assertFalse(any([os.path.exists(f) for f in files]))

        expect = [
            "rm foo.txt",
            "rm bar.txt",
            "rm more.txt",
            "rm even_more.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)

    def test_ssh(self):
        """
        shelephant_rm <sourceinfo.yaml>
        """
        if not has_ssh:
            raise unittest.SkipTest("'ssh localhost' does not work")

        with tempdir(), shelephant.ssh.tempdir("localhost") as remote, contextlib.redirect_stdout(
            io.StringIO()
        ) as sio:
            with cwd(remote):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                shelephant_dump(files)

            shelephant_hostinfo([str(remote), "--ssh", "localhost", "-d"])
            args = [f_hostinfo]
            shelephant_rm(["-n"] + args)
            shelephant_rm(["-f", "--quiet"] + args)
            self.assertFalse(any([os.path.exists(remote / f) for f in files]))

        expect = [
            "rm foo.txt",
            "rm bar.txt",
            "rm more.txt",
            "rm even_more.txt",
        ]
        ret = _plain(sio.getvalue())
        self.assertEqual(ret, expect)


class Test_shelephant_diff(unittest.TestCase):
    def test_output_sha256(self):
        """
        shelephant_diff <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))
                create_dummy_files(["receive.txt"], keep=slice(6, None, None))
                shelephant_dump(["foo.txt", "bar.txt", "receive.txt"])

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                shelephant_dump(["-i"] + files)
                shelephant_hostinfo(["../dest", "-d", "--info"])
                shelephant_diff([f_dump, f_hostinfo, "-o", "foo.yaml"])
                data = shelephant.yaml.read("foo.yaml")

        expect = {
            "!=": ["bar.txt"],
            "->": ["even_more.txt", "more.txt"],
            "<-": ["receive.txt"],
            "==": ["foo.txt"],
        }
        self.assertEqual(data, expect)

    def test_output(self):
        """
        shelephant_diff <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))
                create_dummy_files(["receive.txt"], keep=slice(6, None, None))

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                shelephant_dump(files)
                args = [f_dump, "../dest"]
                if has_rsync:
                    shelephant_diff(args + ["-o", "rsync.yaml", "--mode", "rsync"])
                    data_rsync = shelephant.yaml.read("rsync.yaml")
                shelephant_diff(args + ["-o", "basic.yaml", "--mode", "basic"])
                data_basic = shelephant.yaml.read("basic.yaml")

        if has_rsync:
            expect = {
                "!=": ["bar.txt"],
                "->": ["more.txt", "even_more.txt"],
                "==": ["foo.txt"],
            }
            self.assertEqual(data_rsync, expect)

        expect = {
            "?=": ["foo.txt", "bar.txt"],
            "->": ["more.txt", "even_more.txt"],
        }
        self.assertEqual(data_basic, expect)

    def test_output_filter(self):
        """
        shelephant_diff <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))
                create_dummy_files(["receive.txt"], keep=slice(6, None, None))
                shelephant_dump(["foo.txt", "bar.txt", "receive.txt"])

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                shelephant_dump(["-i"] + files)
                shelephant_hostinfo(["../dest", "-d", "--info"])
                args = [f_dump, f_hostinfo]
                shelephant_diff(args + ["-o", "foo.yaml", "--filter", "<-, ->"])
                data = shelephant.yaml.read("foo.yaml")

        expect = {
            "->": ["even_more.txt", "more.txt"],
            "<-": ["receive.txt"],
        }
        self.assertEqual(data, expect)

    def test_output_filter_list(self):
        """
        shelephant_diff <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir():
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))
                create_dummy_files(["receive.txt"], keep=slice(6, None, None))
                shelephant_dump(["foo.txt", "bar.txt", "receive.txt"])

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                shelephant_dump(["-i"] + files)
                shelephant_hostinfo(["../dest", "-d", "--info"])
                args = [f_dump, f_hostinfo]
                shelephant_diff(args + ["-o", "foo.yaml", "--filter", "<-"])
                data = shelephant.yaml.read("foo.yaml")

        self.assertEqual(data, ["receive.txt"])

    def test_table(self):
        """
        shelephant_diff <sourceinfo.yaml> <destinfo.yaml>
        """
        with tempdir(), contextlib.redirect_stdout(io.StringIO()) as sio:
            pathlib.Path("src").mkdir()
            pathlib.Path("dest").mkdir()

            with cwd("dest"):
                create_dummy_files(["foo.txt"])
                create_dummy_files(["bar.txt"], keep=slice(2, None, None))
                create_dummy_files(["receive.txt"], keep=slice(6, None, None))
                shelephant_dump(["-i", "foo.txt", "bar.txt", "receive.txt"])

            with cwd("src"):
                files = ["foo.txt", "bar.txt", "more.txt", "even_more.txt"]
                create_dummy_files(files)
                shelephant_dump(["-i"] + files)
                shelephant_hostinfo(["../dest", "-d"])
                shelephant_diff([f_dump, f_hostinfo, "--table", "PLAIN_COLUMNS"])

        expect = [
            "bar.txt != bar.txt",
            "even_more.txt ->",
            "more.txt ->",
            "<- receive.txt",
            "foo.txt == foo.txt",
        ]

        ret = _plain(sio.getvalue())[1:]
        self.assertEqual(ret, expect)


if __name__ == "__main__":
    unittest.main()



r"""
Copy & query using *rsync*.

(c) Tom de Geus, 2021, MIT
"""
import os
import re
import subprocess
import tempfile

import numpy as np
import tqdm

from .external import exec_cmd


def copy(
    source_dir: str,
    dest_dir: str,
    files: list[str],
    options: str = "-a",
    verbose: bool = False,
    progress: bool = True,
):
    """
    Copy files using *rsync*.
    This a wrapper around ``rsync {options:s} --files-from``.

    :param source_dir: Source directory. If remote: ``[user@]host:path``.
    :param dest_dir: Source directory. If remote: ``[user@]host:path``.
    :param files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param options: Options passed to ``rsync``.
    :param verbose: Verbose commands.
    :param progress: Show progress bar.
    """

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = os.path.join(temp_dir, "rsync.txt")

        with open(temp_file, "w") as file:
            file.write("\n".join(files))

        # Run without printing output

        if not progress:
            cmd = 'rsync {options:s} --files-from="{files:s}" "{src:s}" "{dest:s}"'.format(
                options=options, src=str(source_dir), dest=str(dest_dir), files=temp_file
            )

            return exec_cmd(cmd, verbose)

        # Run while printing output

        cmd = 'rsync {options:s} -P --files-from="{files:s}" "{src:s}" "{dest:s}"'.format(
            options=options, src=str(source_dir), dest=str(dest_dir), files=temp_file
        )

        if verbose:
            print(cmd)

        pbar = tqdm.tqdm(total=len(files))
        sbar = tqdm.tqdm(unit="B", unit_scale=True)

        process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True)

        for line in iter(process.stdout.readline, b""):
            line = line.decode("utf-8")
            if re.match(r"(.*)(xfe?r\#[0-9])(.*)(to\-che?c?k\=[0-9])(.*)", line):
                e = int(list(filter(None, line.splitlines()[-1].split(" ")))[-6].replace(",", ""))
                pbar.update()
                sbar.update(e)


def diff(
    source_dir: str,
    dest_dir: str,
    files: list[str],
    options: str = "-nai",
    verbose: bool = False,
) -> dict[list[str]]:
    """
    Check if files are different using *rsync*.

    .. note::

        *rsync* uses basic criteria such as file size and creation and modification date.
        This is much faster than using checksums but is only approximate.
        See `rsync manual <https://www.samba.org/ftp/rsync/rsync.html>`_.

    :param str source_dir: Source directory (optionally with hostname).
    :param str dest_dir: Source directory (optionally with hostname).
    :param list files: List of file-paths (relative to ``source_dir`` and ``dest_dir``).
    :param verbose: Verbose commands.
    :return:
        Dictionary with differences::

            {
                "==" : [ ... ], # equal
                "!=" : [ ... ], # not equal
                "->" : [ ... ], # in source_dir not in dest_dir
            }
    """

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file = os.path.join(temp_dir, "rsync.txt")
        files = [os.path.normpath(file) for file in files]

        with open(temp_file, "w") as file:
            file.write("\n".join(files))

        # Run without printing output

        cmd = 'rsync {options:s} --files-from="{files:s}" "{src:s}" "{dest:s}"'.format(
            src=str(source_dir), dest=str(dest_dir), files=temp_file, options=options
        )

        lines = list(filter(None, exec_cmd(cmd, verbose).split("\n")))
        lines = [line for line in lines if line[1] in ["f", "L"]]

        if len(lines) == 0:
            return {
                "==": files,
                "!=": [],
                "->": [],
            }

        check_paths = []
        for line in lines:
            if line[1] == "f":
                check_paths.append(line.split(" ", 1)[1])
            elif line[:2] == "cL":
                check_paths.append(line.split(" ", 1)[1].split(" -> ", 1)[0])

        mode = np.zeros((len(check_paths)), dtype=np.int16)
        modes = {"==": 0, "!=": 1, "->": 2, "<-": 3}

        for i, line in enumerate(lines):
            if line[0] == ">" or line[0] == "<":
                if line[2] == "+":
                    mode[i] = modes["->"]  # create
                else:
                    mode[i] = modes["!="]  # overwrite
            elif line[0] == "c" or line[1] == "L":
                mode[i] = modes["->"]  # create
            elif line[0] == ".":
                pass
            else:
                raise OSError(f'Unknown cryptic output "{line:s}"')

        sorter = np.argsort(files)
        source_paths = np.array(files, dtype=str)[sorter]

        i = np.argsort(check_paths)
        check_paths = np.array(check_paths, dtype=str)[i]
        mode = mode[i]

        test = np.in1d(source_paths, check_paths)

        idx = np.searchsorted(check_paths, source_paths)
        idx = np.where(test, idx, 0)
        ret = np.where(test, mode[idx], 0)
        ret = ret.astype(np.int16)
        out = np.empty_like(ret)
        out[sorter] = ret

        files = np.array(files)

        return {
            "==": files[out == modes["=="]].tolist(),
            "->": files[out == modes["->"]].tolist(),
            "!=": files[out == modes["!="]].tolist(),
        }



import json
import os
import pathlib
import re
import subprocess
import sys
import tempfile
from contextlib import contextmanager


@contextmanager
def tempdir():
    """
    Set the cwd to a temporary directory::

        with tempdir("foo"):
            # Do something in foo
    """

    origin = pathlib.Path().absolute()
    with tempfile.TemporaryDirectory() as dirname:
        try:
            os.chdir(dirname)
            yield
        finally:
            os.chdir(origin)


@contextmanager
def cwd(dirname: pathlib.Path):
    """
    Set the cwd to a specified directory::

        with cwd("foo"):
            # Do something in foo

    :param dirname: The directory to change to.
    """
    origin = pathlib.Path().absolute()
    try:
        os.chdir(dirname)
        yield
    finally:
        os.chdir(origin)


def _check_skip(path: str, skip: list[str]) -> bool:
    """
    Check if a path should be skipped.

    :param path: The path to check.
    :param skip: A list of regex to skip.
    :return: ``True`` if the path should be skipped.
    """
    for s in skip:
        if re.match(s, path):
            return True
    return False


def _search_rglob(rglob: str, skip: list[str] = []) -> list[pathlib.Path]:
    """
    Search for files using ``rglob``.

    :param rglob: The pattern to search for.
    :param skip: A list of regex to skip.
    :return: A list of paths.
    """
    if isinstance(skip, str):
        skip = [skip]
    ret = []
    for path in pathlib.Path(".").rglob(rglob):
        if _check_skip(str(path), skip):
            continue
        ret.append(path)
    return ret


def _search_glob(glob: str, skip: list[str] = []) -> list[pathlib.Path]:
    """
    Search for files using ``glob``.

    :param glob: The pattern to search for.
    :param skip: A list of regex to skip.
    :return: A list of paths.
    """
    if isinstance(skip, str):
        skip = [skip]
    ret = []
    for path in pathlib.Path(".").glob(glob):
        if _check_skip(str(path), skip):
            continue
        ret.append(path)
    return ret


def _search_exec(exec: str, skip: list[str] = []) -> list[pathlib.Path]:
    """
    Search for files using a command.

    :param exec: The command to run.
    :param skip: A list of regex to skip.
    :return: A list of paths.
    """
    if isinstance(skip, str):
        skip = [skip]
    paths = subprocess.check_output(exec, shell=True).decode("utf-8").splitlines()
    paths = list(filter(None, paths))
    ret = []
    for path in paths:
        if _check_skip(path, skip):
            continue
        ret.append(path)
    return list(map(pathlib.Path, map(os.path.normpath, ret)))


def search(*settings: dict, root: pathlib.Path = pathlib.Path(".")) -> list[pathlib.Path]:
    r"""
    Search for files using a list of settings, as follows::

        [
            {"rglob": "*.py", "skip": ["\..*", "build"]},
            {"exec": "find . -name '*.cpp'"},
        ]

    :param settings: A list of settings.
    :param root: The root directory to search in.
    :return: A list of paths.
    """
    with cwd(root):
        ret = []
        for setting in settings:
            if "rglob" in setting:
                ret += _search_rglob(**setting)
            elif "exec" in setting:
                ret += _search_exec(**setting)
            elif "glob" in setting:
                ret += _search_glob(**setting)
            else:
                raise ValueError(f"Unknown search setting: {setting}")
        return ret


if __name__ == "__main__":
    settings = json.loads(pathlib.Path("settings.json").read_text())
    root = sys.argv[1] if len(sys.argv) > 1 else pathlib.Path(".")
    files = search(*settings, root=pathlib.Path(root))
    pathlib.Path("files.txt").write_text("\n".join(map(str, files)))
